from __future__ import annotations

from typing import TYPE_CHECKING, Iterable
import numpy as np

from ..utils.trackable import Store
from ..resources import Resource, Buffer, Texture

if TYPE_CHECKING:
    from numpy.typing import ArrayLike

    ArrayOrBuffer = ArrayLike | Buffer
    ArrayOrTexture = ArrayLike | Texture
    ArrayOrTextureOrBuffer = ArrayLike | Texture | Buffer

Optional = None


class Geometry(Store):
    """An object's geometry is a container for data that 'defines' (the shape of) the object.

    This class has no documented properties; attributes (usually buffers and
    sometimes textures) can be freely added to it. What attributes are required
    depends on the kind of object and the usage of the material. Some attributes
    are optional. However, there are several common names.

    Common buffer attributes:

    * ``positions``: an Nx3 buffer representing vertex positions. Used by e.g. ``Mesh``, ``Line``, ``Points``, ``Text``.
    * ``indices``: an Nx3 buffer representing the triangular faces of a ``Mesh``.
    * ``normals``: an Nx3 buffer representing the surface normals of a ``Mesh``.
    * ``texcoords``: an Nx1, Nx2, or Nx3 set of per-vertex texture coordinates. The dimensionality
      should match that of the dimension of the colormap's texture (``material.map``).
    * ``texcoords1``, ``texcoords2`` etc.: for additional texture coordinates. Usually Nx2.
      E.g. a ``TextureMap`` with ``uv_channel`` set to 4 will use "texcoords4".
    * ``colors``: per vertex or per-face color data for e.g. ``Mesh``, ``Line``, ``Points``.
      Can be Nx1 (grayscale), Nx2 (gray plus alpha), Nx3 (RGB), or Nx4 (RGBA).
    * ``sizes``: per vertex sizes for e.g. ``Points``.
    * ``edge_colors`` per vertex edge colors for points with the marker material.
    * ``rotations``: per vertex point/marker rotations.

    Common texture attributes:

    * ``grid``: a 2D or 3D texture for the ``Image`` and ``Volume`` objects, respectively.

    For mesh morphing the following attributes are used:

    * ``morph_targets_relative``: a bool indicating whether the morph positions are relative.
    * ``morph_positions``: a list of arrays of per-vertex morph positions.
    * ``morph_normals``: a list of arrays of per-vertex morph normals.
    * ``morph_colors``: a list of arrays of per-vertex morph colors.

    Instantiation
    -------------
    Most attributes of the geometry are buffers or textures. For convenience, these
    can be passed as arrays, in which case they are automatically wrapped in a buffer
    or texture.

    Example
    -------

    .. code-block:: py

        g = Geometry(positions=[[1, 2], [2, 4], [3, 5], [4, 1]])
        g.positions  # Buffer
        g.positions.data  # numpy array

    """

    def __init__(
        self,
        *,
        positions: ArrayOrBuffer = Optional,
        indices: ArrayOrBuffer = Optional,
        normals: ArrayOrBuffer = Optional,
        texcoords: ArrayOrBuffer = Optional,
        colors: ArrayOrBuffer = Optional,
        grid: ArrayOrTexture = Optional,
        **other_attributes: ArrayOrTextureOrBuffer,
    ):
        super().__init__()

        # We separately declare some possible attribute in the signature to help users via autocompletion.
        # We just merge these with the kwargs so we can process them uniformly.
        all_attributes = other_attributes.copy()
        if positions is not Optional:
            all_attributes["positions"] = positions
        if indices is not Optional:
            all_attributes["indices"] = indices
        if normals is not Optional:
            all_attributes["normals"] = normals
        if texcoords is not Optional:
            all_attributes["texcoords"] = texcoords
        if colors is not Optional:
            all_attributes["colors"] = colors
        if grid is not Optional:
            all_attributes["grid"] = grid

        for name, val in all_attributes.items():
            # Get resource object
            if isinstance(val, Resource):
                resource = val
            else:
                # Convert literal arrays to numpy arrays (buffers and textures require memoryview compatible data).
                if isinstance(val, list):
                    dtype = "int32" if name == "indices" else "float32"
                    val = np.array(val, dtype=dtype)
                # Create texture or buffer
                if name == "grid":
                    val = np.asanyarray(val)
                    dim = val.ndim
                    if dim > 2 and val.shape[-1] <= 4:
                        dim -= 1  # last array dim is probably (a subset of) rgba
                    resource = Texture(val, dim=dim)
                else:
                    resource = Buffer(val)

            # Checks
            if isinstance(resource, Buffer):
                format = resource.format
                if name == "indices":
                    # Make no assumptions about shape. Shader will need to validate.
                    # For meshes will be Nx3 or Nx4, but other dtypes may support
                    # multidimensional stuff for fancy graphics.
                    pass
                elif format is None:
                    pass  # buffer with no local data, trust the user
                elif name == "positions":
                    if not format.startswith("3x"):
                        raise ValueError("Expected Nx3 data for positions")
                elif name == "normals":
                    if not format.startswith("3x"):
                        raise ValueError("Expected Nx3 data for normals")
                elif name == "texcoords":
                    if not ("x" not in format or format.startswith(("1x", "2x", "3x"))):
                        raise ValueError("Expected Nx1, Nx2 or Nx3 data for texcoords")
                elif name == "colors":
                    if not format.startswith(("3x", "4x")):
                        raise ValueError("Expected Nx3 or Nx4 data for colors")
                elif name == "sizes":
                    if not ("x" not in format):
                        raise ValueError("Expected array of scalars for sizes")
                else:
                    pass  # Unknown attribute - no checks

            # Store
            setattr(self, name, resource)

    def __repr__(self) -> str:
        # A Store is a subclass of a dict, but it does not look like a dict,
        # e.g. you cannot do geometry.items() or any of the regular dict
        # methods, because *all*  atrribute access is converted to dict key
        # access. So let's forget about this being a dict and also provide a
        # useful repr.
        lines = ["Geometry("]
        for key in dir(self):
            val = self[key]
            lines.append(f"    {key}={val!r},")
        lines.append(f") # at {hex(id(self))}")
        return "\n".join(lines)

    def __dir__(self) -> Iterable[str]:
        return sorted([name for name in self if not name.startswith("_trackable_")])



import numpy as np

from .. import Geometry, Line, LineSegmentMaterial
from ..utils import Color

DTYPE = "f4"


class GridHelper(Line):
    """A WorldObject that shows a grid-shaped wireframe.

    The generated grid will be in the xz-plane centered at the origin. To
    position the grid, manipulate its parent's position, rotation, etc.

    Parameters
    ----------
    size : float
        The size of the wireframe in the direction of the x-, and y-axis.
    divisions : int
        The number of (evenly spaced) divisions to perform along each axis.
    color1 : int, float, str, tuple
        The color of the center lines. This is a either a single int or float
        (gray), a 4-tuple ``(r,g,b,a)`` of ints or floats, or a hex-coded color
        string in one of the following formats: ``#RGB``, ``#RGBA``,
        ``#RRGGBB``, ``#RRGGBBAA``.
    color2 : int, float, str, tuple
        The color of non-center lines. This is a either a single int or float
        (gray), a 4-tuple ``(r,g,b,a)`` of ints or floats, or a hex-coded color
        string in one of the following formats: ``#RGB``, ``#RGBA``,
        ``#RRGGBB``, ``#RRGGBBAA``.
    thickness : int
        The thickness in screen units (pixels).

    """

    def __init__(
        self,
        size=10.0,
        divisions=10,
        color1=(0.35, 0.35, 0.35, 1),
        color2=(0.1, 0.1, 0.1, 1),
        thickness=1,
    ):
        assert isinstance(divisions, int)
        assert size > 0.0

        half_size = size / 2
        n_lines = divisions + 1
        x = np.linspace(-half_size, half_size, num=n_lines, dtype=DTYPE)

        # the grid is made up of 2 * n_lines line segments
        # where each line has two endpoints (2, 3)
        positions = np.zeros((2, n_lines, 2, 3), dtype=DTYPE)
        positions[0, ..., 0] = x[:, None]
        positions[0, ..., 2] = [[-half_size, half_size]]
        positions[1, ..., 0] = [[-half_size, half_size]]
        positions[1, ..., 2] = x[:, None]

        # color1 for the center lines, color2 for the rest
        colors = np.empty((2, n_lines, 2, 4), dtype=DTYPE)
        colors[..., :] = Color(color2)
        colors[:, n_lines // 2, :, :] = Color(color1)

        geometry = Geometry(
            positions=positions.reshape((-1, 3)), colors=colors.reshape((-1, 4))
        )
        material = LineSegmentMaterial(
            color_mode="vertex", thickness=thickness, aa=True
        )

        super().__init__(geometry, material)



"""
Containers for buffers and textures.

In Pygfx, data is stored in buffers and textures. We collectively call these resources.

.. currentmodule:: pygfx.resources

.. autosummary::
    :toctree: resources/
    :template: ../_templates/custom_layout.rst

    Resource
    Buffer
    Texture
    TextureMap

"""

from ._base import Resource
from ._buffer import Buffer
from ._texture import Texture
from ._texturemap import TextureMap

__all__ = ["Buffer", "Resource", "Texture", "TextureMap"]



from ._perspective import PerspectiveCamera


class OrthographicCamera(PerspectiveCamera):
    """An orthographic camera, useful for orthograpic views and 2D content.

    Technically, this is a PerspectiveCamera with the fov locked to zero.
    You can use this instead of a PerspectiveCamera if you have data that does
    not make sense to view with a nonzero fov.

    Parameters
    ----------
    width: float
        The (minimum) width of the view-cube. The actual view
        may be wider if the viewport is relatively wide. Default 1.
    height: float
        The (minimum) height of the view-cube. The actual view
        may be taller if the viewport is relatively high. Default 1.
    zoom: float
        An additional zoom factor, equivalent to attaching a zoom lens.
    maintain_aspect: bool
        Whether the aspect ration is maintained as the window size changes.
        Default True. If false, the dimensions are stretched to fit the window.
    depth: float | None
        The reference size of the scene in the depth dimension. By default this value gets set
        to the average of ``width`` and ``height``. This happens on initialization and by ``show_pos()``, ``show_object()`` and ``show_rect()``.
        It is used to calculate a sensible depth range when ``depth_range`` is not set.
    depth_range: 2-tuple | None
        The explicit values for the near and far clip planes. If not given
        or None (the default), the clip planes ware calculated using ``fov`` and ``depth``.

    """

    _fov_range = 0, 0  # Disallow fov other than zero

    def __init__(
        self,
        width=1,
        height=1,
        *,
        zoom=1,
        maintain_aspect=True,
        depth=None,
        depth_range=None,
    ):
        super().__init__(
            0,
            None,
            width=width,
            height=height,
            zoom=zoom,
            maintain_aspect=maintain_aspect,
            depth=depth,
            depth_range=depth_range,
        )



import numpy as np
from ..utils.trackable import Trackable
from ..resources import Buffer
from ..utils import array_from_shadertype


class TextureMap(Trackable):
    """A texture map.

    TextureMap is used to define how a texture should be used in a material.
    including which uv channel to use, filtering, and wrapping

    Parameters
    ----------
    texture : Texture
        The texture to use for this map.
    uv_channel : int, optional
        The uv channel of the texture to use.
        e.g. with `channel=7`, it will use the `geometry.texcoords7`.
        Default is 0, which uses the `geometry.texcoords`.
    filter : str, optional
        The filter to use for magnification, minification znd mipmap, if not being set separately.
        Can be "nearest" or "linear". Default is "linear".
    mag_filter : str, optional
        The magnification filter to use. Defaults to the value of ``filter``.
    min_filter : str, optional
        The minification filter to use. Defaults to the value of ``filter``.
    mipmap_filter : str, optional
        The mipmap filter to use. Defaults to the value of ``filter``.
    wrap : str, optional
        The wrap mode for both the s and t coordinates if not being set separately.
        Can be "clamp", "repeat", "mirror". Default is "repeat".
    wrap_s : str, optional
        The wrap mode for the s coordinate. Defaults to the value of ``wrap``.
    wrap_t : str, optional
        The wrap mode for the t coordinate. Defaults to the value of ``wrap``.
    """

    uniform_type = dict(
        transform="3x3xf4",
    )

    def __init__(
        self,
        texture,
        *,
        uv_channel=0,
        filter="linear",
        mag_filter=None,
        min_filter=None,
        mipmap_filter=None,
        wrap="repeat",
        wrap_s=None,
        wrap_t=None,
    ):
        super().__init__()
        self.texture = texture
        self.uv_channel = uv_channel
        self.mag_filter = mag_filter or filter
        self.min_filter = min_filter or filter
        self.mipmap_filter = mipmap_filter or filter
        self.wrap_s = wrap_s or wrap
        self.wrap_t = wrap_t or wrap

        self._offset = (0, 0)
        self._scale = (1, 1)
        self._rotation = 0

        self._store.uniform_buffer = Buffer(
            array_from_shadertype(self.uniform_type), force_contiguous=True
        )
        self.update_matrix()

    @property
    def texture(self):
        """The texture to use for this map."""
        return self._store.texture

    @texture.setter
    def texture(self, value):
        self._store.texture = value

    @property
    def uv_channel(self):
        """The channel of the texture to use."""
        return self._store.uv_channel

    @uv_channel.setter
    def uv_channel(self, value):
        self._store.uv_channel = value

    @property
    def mag_filter(self):
        """The magnification filter to use."""
        return self._store.mag_filter

    @mag_filter.setter
    def mag_filter(self, value):
        self._store.mag_filter = value

    @property
    def min_filter(self):
        """The minification filter to use."""
        return self._store.min_filter

    @min_filter.setter
    def min_filter(self, value):
        self._store.min_filter = value

    @property
    def mipmap_filter(self):
        """The mipmap filter to use."""
        return self._store.mipmap_filter

    @mipmap_filter.setter
    def mipmap_filter(self, value):
        self._store.mipmap_filter = value

    @property
    def wrap_s(self):
        """The wrap mode for the s coordinate."""
        return self._store.wrap_s

    @wrap_s.setter
    def wrap_s(self, value):
        self._store.wrap_s = value

    @property
    def wrap_t(self):
        """The wrap mode for the t coordinate."""
        return self._store.wrap_t

    @wrap_t.setter
    def wrap_t(self, value):
        self._store.wrap_t = value

    @property
    def offset(self):
        """The uv offset of the texture map.

        Note: You should call `update_matrix` after changing this value to update the matrix.
        """
        return self._offset

    @offset.setter
    def offset(self, value):
        self._offset = value

    @property
    def scale(self):
        """The uv scale of the texture map.

        Note: You should call `update_matrix` after changing this value to update the matrix.
        """
        return self._scale

    @scale.setter
    def scale(self, value):
        self._scale = value

    @property
    def rotation(self):
        """The rotation of the texture map in uv plane.

        Note: You should call `update_matrix` after changing this value to update the matrix.
        """
        return self._rotation

    @rotation.setter
    def rotation(self, value):
        self._rotation = value

    def update_matrix_components(self, offset=None, scale=None, rotation=None):
        """Update the texture map's matrix with the given offset, scale and rotation.

        Parameters
        ----------
        offset : tuple of float, optional
            The offset of the texture map.
        scale : tuple of float, optional
            The scale of the texture map.
        rotation : float, optional
            The rotation of the texture map.
        """

        if offset is not None:
            self._offset = offset

        if scale is not None:
            self._scale = scale

        if rotation is not None:
            self._rotation = rotation

        self.update_matrix()

    # Currently, we let users to manage the matrix update timing themselves.
    # todo: auto update matrix when offset, scale, rotation are changed before the next draw call.
    def update_matrix(self):
        """Update the matrix of the texture map."""

        cos_r = np.cos(self._rotation)
        sin_r = np.sin(self._rotation)

        matrix = np.array(
            [
                [cos_r * self._scale[0], sin_r * self._scale[0], self._offset[0]],
                [-sin_r * self._scale[1], cos_r * self._scale[1], self._offset[1]],
                [0, 0, 1],
            ]
        )

        self.uniform_buffer.data["transform"][:, :3] = matrix.T
        self.uniform_buffer.update_full()

    @property
    def uniform_buffer(self):
        """The uniform buffer of the texture map."""
        return self._store.uniform_buffer



import wgpu

from .... import objects
from .update import ensure_wgpu_object, update_resource
from .utils import to_index_format, to_vertex_format, GpuCache
from .shared import get_shared


# This cache enables re-using gpu pipelines for calculating shadows,
# these can be shared between multiple world-objects that have a
# positions buffer with a matching stride and format.
SHADOW_CACHE = GpuCache("shadow_pipelines")


shadow_vertex_shader = """
    struct Matrix4x4 {
        m : mat4x4<f32>
    };
    @group(0) @binding(0) var<uniform> light_view_projection : Matrix4x4;
    @group(1) @binding(0) var<uniform> model_transform : Matrix4x4;
    @vertex
    fn vs_main(@location(0) position: vec3<f32>) -> @builtin(position) vec4<f32> {
        return light_view_projection.m * model_transform.m * vec4<f32>(position, 1.0);
    }
"""

instanced_shadow_vertex_shader = """
    struct Matrix4x4 {
        m : mat4x4<f32>
    };

    struct InstanceInput {
        @location(1) model_matrix_0: vec4<f32>,
        @location(2) model_matrix_1: vec4<f32>,
        @location(3) model_matrix_2: vec4<f32>,
        @location(4) model_matrix_3: vec4<f32>,
    };

    @group(0) @binding(0) var<uniform> light_view_projection : Matrix4x4;
    @group(1) @binding(0) var<uniform> model_transform : Matrix4x4;
    @vertex
    fn vs_main(@location(0) position: vec3<f32>, instance: InstanceInput) -> @builtin(position) vec4<f32> {
        let instance_matrix = mat4x4<f32>(
            instance.model_matrix_0,
            instance.model_matrix_1,
            instance.model_matrix_2,
            instance.model_matrix_3
        );
        return light_view_projection.m * (model_transform.m * instance_matrix) * vec4<f32>(position, 1.0);
    }
"""

bind_group_layout_entry = {
    "binding": 0,
    "visibility": wgpu.ShaderStage.VERTEX,
    "buffer": {"type": wgpu.BufferBindingType.uniform},
}


global_bind_group_layout = None


def render_shadow_maps(lights, wobjects, command_encoder):
    """Render the wobjects into the shadow maps for the given lights."""

    global global_bind_group_layout

    # Make sure the global bind_group_layout is ready
    device = get_shared().device
    if global_bind_group_layout is None:
        global_bind_group_layout = device.create_bind_group_layout(
            entries=[bind_group_layout_entry]
        )

    # Filter shadow-able objects once beforehand.
    # We assume this has already been done
    # wobjects = [w for w in wobjects if w.cast_shadow and w.geometry is not None]

    for light in lights:
        if not light.cast_shadow:
            continue

        # What kind of light is this?
        light_has_6_sides = isinstance(light, objects.PointLight)
        assert light_has_6_sides == isinstance(light.shadow._wgpu_tex_view, list)
        assert light_has_6_sides == isinstance(light.shadow._gfx_matrix_buffer, list)

        if light_has_6_sides:
            # Render each shadow map
            shadow_maps = light.shadow._wgpu_tex_view
            shadow_buffers = light.shadow._gfx_matrix_buffer
            for i in range(6):
                render_shadow_map(
                    device,
                    light,
                    shadow_maps[i],
                    shadow_buffers[i],
                    wobjects,
                    command_encoder,
                )
        else:
            # Render this one shadow map
            shadow_map = light.shadow._wgpu_tex_view
            shadow_buffer = light.shadow._gfx_matrix_buffer
            render_shadow_map(
                device, light, shadow_map, shadow_buffer, wobjects, command_encoder
            )


def render_shadow_map(
    device, light, shadow_map, shadow_buffer, wobjects, command_encoder
):
    """Render the wobjects into the given shadow map."""

    shadow_pass = command_encoder.begin_render_pass(
        color_attachments=[],
        depth_stencil_attachment={
            "view": shadow_map,
            "depth_read_only": False,
            "depth_clear_value": 1.0,
            "depth_load_op": wgpu.LoadOp.clear,
            "depth_store_op": wgpu.StoreOp.store,
            # The used texture does not have a stencil component
            # "stencil_read_only": True,
            # "stencil_load_op": wgpu.LoadOp.clear,
            # "stencil_store_op": wgpu.StoreOp.discard,
        },
    )

    light_bind_group = get_shadow_bind_group(device, shadow_buffer)
    shadow_pass.set_bind_group(0, light_bind_group, [], 0, 99)

    for wobject in wobjects:
        render_wobject_shadow(device, light, wobject, shadow_pass)

    shadow_pass.end()


def render_wobject_shadow(device, light, wobject, shadow_pass):
    """Render one wobject into a shadow map."""

    shadow_pipeline = get_shadow_pipeline(device, wobject, light.shadow.cull_mode)
    shadow_pass.set_pipeline(shadow_pipeline)

    position_buffer = wobject.geometry.positions
    shadow_pass.set_vertex_buffer(
        0,
        ensure_wgpu_object(position_buffer),
        position_buffer.draw_range[0] * position_buffer.itemsize,
        position_buffer.draw_range[1] * position_buffer.itemsize,
    )

    n_instance = 1

    if isinstance(wobject, objects.InstancedMesh):
        instance_buffer = wobject.instance_buffer
        shadow_pass.set_vertex_buffer(1, ensure_wgpu_object(instance_buffer))
        n_instance = instance_buffer.nitems

    wobject_bind_group = get_shadow_bind_group(device, wobject.uniform_buffer)
    shadow_pass.set_bind_group(1, wobject_bind_group, [], 0, 99)

    ibuffer = getattr(wobject.geometry, "indices", None)

    if ibuffer is not None:
        n = wobject.geometry.indices.data.size
        shadow_pass.set_index_buffer(
            ensure_wgpu_object(ibuffer), to_index_format(ibuffer.format)
        )
        shadow_pass.draw_indexed(n, n_instance)
    else:
        n = wobject.geometry.positions.nitems
        shadow_pass.draw(n, n_instance)


def get_shadow_bind_group(device, shadow_buffer):
    """Get the bind group object for this shadow buffer."""

    # Since the bind-group is bound one-to-one to the buffer, there is
    # no need for sophisticated caching:  The bind group is cached on
    # the buffer itself.

    bind_group = getattr(shadow_buffer, "_gfx_shadow_bind_group", None)

    if bind_group is None:
        shadow_buffer._wgpu_usage |= wgpu.BufferUsage.UNIFORM
        wgpu_buffer = ensure_wgpu_object(shadow_buffer)
        # We also update the buffer now, because this code gets called
        # *after* the renderer has synced all resources. Note that this
        # code is only reached once for each shadow map.
        update_resource(shadow_buffer)

        bind_group = device.create_bind_group(
            layout=global_bind_group_layout,
            entries=[
                {
                    "binding": 0,
                    "resource": {
                        "buffer": wgpu_buffer,
                        "offset": 0,
                        "size": 64,
                    },
                }
            ],
        )

        shadow_buffer._gfx_shadow_bind_group = bind_group

    return bind_group


def get_shadow_pipeline(device, wobject, cull_mode):
    """Get the pipeline object for rendering a wobject into a shadow map."""

    # The shadow pipeline only depends on the object's geometry info
    # (Vertex Buffer), so there is potential for sharing this pipeline
    # between objects. We use a proper cache to implement this.
    #
    # NOTE:
    # - for instanced meshes, it also needs instance buffer.
    # - maybe for future skinned meshes, the morph buffer is also needed.

    position_buffer = wobject.geometry.positions

    stride = position_buffer.itemsize
    format = position_buffer.format
    topology = get_shadow_topology(wobject)

    is_instanced = isinstance(wobject, objects.InstancedMesh)

    key = (stride, format, topology, cull_mode, is_instanced)
    wobject_attr_name = f"_gfx_shadow_pipeline_{cull_mode}"

    # Try getting the pipeline from the wobject
    try:
        current_key, pipeline = getattr(wobject, wobject_attr_name)
    except AttributeError:
        current_key, pipeline = None, None
    if current_key != key:
        pipeline = None

    # If not existing or compatible, try getting from caache (i.e. re-use from other object)
    if pipeline is None:
        pipeline = SHADOW_CACHE.get(key)

    # If still nothing, create the pipeline and make available for re-use.
    if pipeline is None:
        # Create pipeline and store in the cache (with a weakref).
        pipeline = create_shadow_pipeline(
            device, stride, format, topology, cull_mode, is_instanced
        )
        SHADOW_CACHE.set(key, pipeline)

    # Store on the wobject to bind it to its lifetime, but per shadow-cull-mode
    setattr(wobject, wobject_attr_name, (key, pipeline))
    return pipeline


def get_shadow_topology(wobject):
    """Get the primitive topology for drawing the shadows of this object."""
    if isinstance(wobject, objects.Mesh):
        return wgpu.PrimitiveTopology.triangle_list
    elif isinstance(wobject, objects.Line):
        return wgpu.PrimitiveTopology.line_strip
    elif isinstance(wobject, objects.Points):
        return wgpu.PrimitiveTopology.point_list
    else:
        raise RuntimeError(f"Shadows not supported for {wobject.__class__.__name__}")


def create_shadow_pipeline(
    device, stride, format, topology, cull_mode, instanced=False
):
    """Actually create a shadow pipeline object."""

    vertex_buffer_descriptor = [
        {
            "array_stride": stride,
            "step_mode": wgpu.VertexStepMode.vertex,  # vertex
            "attributes": [
                {
                    "format": to_vertex_format(format),
                    "offset": 0,
                    "shader_location": 0,
                }
            ],
        }
    ]

    if instanced:
        vertex_buffer_descriptor.append(
            {
                "array_stride": 80,  # matrix4x4(64) + id(4) + padding(12) = 80
                "step_mode": wgpu.VertexStepMode.instance,  # instance
                "attributes": [
                    {
                        "format": "float32x4",
                        "offset": 0,
                        "shader_location": 1,
                    },
                    {
                        "format": "float32x4",
                        "offset": 16,
                        "shader_location": 2,
                    },
                    {
                        "format": "float32x4",
                        "offset": 32,
                        "shader_location": 3,
                    },
                    {
                        "format": "float32x4",
                        "offset": 48,
                        "shader_location": 4,
                    },
                ],
            }
        )
        shader_module = device.create_shader_module(code=instanced_shadow_vertex_shader)
    else:
        shader_module = device.create_shader_module(code=shadow_vertex_shader)

    pipeline = device.create_render_pipeline(
        layout=device.create_pipeline_layout(
            bind_group_layouts=[global_bind_group_layout, global_bind_group_layout]
        ),
        vertex={
            "module": shader_module,
            "entry_point": "vs_main",
            "buffers": vertex_buffer_descriptor,
        },
        primitive={
            "topology": topology,
            "cull_mode": cull_mode.lower(),
        },
        depth_stencil={
            "format": wgpu.TextureFormat.depth32float,
            "depth_write_enabled": True,
            "depth_compare": wgpu.CompareFunction.less,
            "stencil_read_mask": 0,
            "stencil_write_mask": 0,
        },
    )

    return pipeline



import time

from .. import (
    plane_geometry,
    Mesh,
    Group,
    Text,
    TextMaterial,
    ScreenCoordsCamera,
    MeshBasicMaterial,
)


class Stats(Group):
    """A Stats helper which displays performance statistics such
    as FPS and draw time on the screen.

    Parameters
    ----------
    viewport : Viewport, Renderer
        Required for positioning lines of text in the corner of
        the screen.

    """

    def __init__(
        self,
        viewport,
    ):
        from .. import Viewport  # avoid circular import

        super().__init__()

        self._line_height = 16
        font_size = 12
        foreground = "#0f0"
        background = "#020"
        text_material = TextMaterial(color=foreground)

        quad_geometry = plane_geometry()
        quad_geometry.positions.data[..., :2] += [0.5, -0.5]

        self.bg = Mesh(
            quad_geometry,
            MeshBasicMaterial(
                color=background,
                side="both",
                opacity=0.9,
            ),
        )
        # refactor once text bounding boxes are available
        self.bg.local.scale = (90, self._line_height * 2.1, 1)
        self.stats_text = Text(
            material=text_material,
            text="",
            screen_space=True,
            font_size=font_size,
            anchor="topleft",
        )
        self.add(self.bg, self.stats_text)

        self.camera = ScreenCoordsCamera()

        # track screen size for line positioning
        self._viewport = Viewport.from_viewport_or_renderer(viewport)
        self._update_positions()
        self._viewport.renderer.add_event_handler(
            self._update_positions,
            "resize",
        )

        # flag used to skip the first frame
        # which typically has all the CPU->GPU transfer and
        # shader compilation overhead
        self._init = False

        # performance trackers
        self._tmin = 1e10
        self._tmax = 0
        self._tbegin = None
        self._tprev = self._tbegin
        self._frames = 0
        self._fmin = 1e10
        self._fmax = 0
        # Sentinel value of None indicates that the fps has never been computed
        self._fps = None

    def _update_positions(self, event=None):
        _, height = self._viewport.logical_size
        self.stats_text.local.position = (0, height, 0)
        self.bg.local.position = (0, height, 0.1)

    def start(self):
        if not self._init:
            return

        self._tbegin = time.perf_counter_ns()
        if self._tprev is None:
            self._tprev = self._tbegin

    def stop(self):
        if not self._init:
            self._init = True
            return

        t = time.perf_counter_ns()
        self._frames += 1

        delta = round((t - self._tbegin) / 1_000_000)
        self._tmin = min(self._tmin, delta)
        self._tmax = max(self._tmax, delta)

        if t >= self._tprev + 1_000_000_000:
            # update FPS counter whenever a second has passed
            fps = round(self._frames / ((t - self._tprev) / 1_000_000_000))
            self._tprev = t
            self._frames = 0
            self._fmin = min(self._fmin, fps)
            self._fmax = max(self._fmax, fps)
            self._fps = fps

        text = f"{delta} ms ({self._tmin}-{self._tmax})"
        if self._fps is not None:
            text += f"\n{self._fps} fps ({self._fmin}-{self._fmax})"
        self.stats_text.set_text(text)

    def render(self, flush=True):
        self._viewport.render(self, self.camera, flush=flush)

    def __enter__(self):
        self.start()

    def __exit__(self, *exc):
        self.stop()



import numpy as np
import freetype

from ._atlas import glyph_atlas
from ._shaper import CACHE_FT


# A little cache so we can assign numbers to fonts
fontname_cache = {}


def generate_glyph(glyph_indices, font_filename):
    """Generate a glyph for the given glyph indices.

    Parameters:
        glyph_indices (list): the indices in the font to render a glyph for.
        font_filename (str): the font to use.

    This generates SDF glyphs and puts them in the atlas. The indices
    of where the glyphs are in the atlas are returned. Glyphs already
    present in the atlas are reused.
    """

    # The glyph metrics place the origin at the baseline. When a bitmap
    # is generated, the bitmap's origin is not the glyphs origin, so
    # we need to correct for this, otherwise the characters are not on
    # the same baseline. Further, the glyph bitmap that is generated
    # varies in size depending on the glyph. The rect info to account for
    # the above points is stored in the per-glyph buffer of the atlas.

    # Get font index (so we can make it part of the glyph hash)
    try:
        font_index = fontname_cache[font_filename]
    except KeyError:
        font_index = len(fontname_cache) + 1
        fontname_cache[font_filename] = font_index

    # Get the face object. We will not need it if all glyphs are already
    # in the atlas, but because of the cache it is fast, and this way
    # we keep the face alive in the cache.
    face = CACHE_FT[font_filename]

    atlas_indices = np.empty((len(glyph_indices),), "u4")
    for i in range(len(glyph_indices)):
        glyph_index = int(glyph_indices[i])
        glyph_hash = (font_index, glyph_index)
        index = glyph_atlas.get_index_from_hash(glyph_hash)
        if index is None:
            glyph, offset = _generate_sdf(face, glyph_index)
            index = glyph_atlas.store_region_with_hash(glyph_hash, glyph, offset)
        atlas_indices[i] = index

    return atlas_indices


def _generate_sdf(face, glyph_index):
    """Generate the SDF bitmap."""
    # This only gets called for glyphs that are not in the atlas yet.

    # FreeType has two ways to render an SDF. The old way is based on
    # the bitmap, and the new way renders from the geometry directly.
    # The new way is the default, and the old way can be selected by
    # first rendering the glyph in bitmap mode. See
    # http://freetype.org/freetype2/docs/reference/ft2-base_interface.html
    #
    # Currently the new way produces quite a few artifacts in e.g. Arabic
    # and emoticons, due to sharp and/or intersecting Bezier curves.
    #
    # Some timing data:
    # about 0.2 ms to create a simple bitmap.
    # about 0.8 ms to create an SDF from the bitmap in the old/stable way (no artifacts)
    # about 1.9 ms to create an SDF the new/fancy way (with artifacts)
    #
    # So, at the time I write this (November 2022) the SDF that goes
    # via the bitmap is both more stable and faster. We might want to
    # revisit this in a year or so.

    # Load the glyph bitmap
    face.load_glyph(glyph_index, freetype.FT_LOAD_DEFAULT)

    # Render bitmap. This forces the SDF to be based off the bitmap, being more stable
    face.glyph.render(freetype.FT_RENDER_MODE_NORMAL)

    # Render SDF
    try:
        face.glyph.render(freetype.FT_RENDER_MODE_SDF)
    except Exception:  # Freetype refuses SDF for spaces ?
        pass

    # Convert to numpy array
    bitmap = face.glyph.bitmap
    glyph = np.array(bitmap.buffer, np.uint8).reshape(bitmap.rows, bitmap.width)
    offset = face.glyph.bitmap_left, face.glyph.bitmap_top

    return glyph, offset



import re


matchers = {
    "nl": r"(\n)+",  # newline (nl) -- match first so that it doesn't get included in \s
    "ws": r"([ \r\f\t])+",  # whitespace (ws) (except for newline as it is matched above)
    "word": r"(\w)+",
    "number": r"(\d)+",
    "punctuation": r"([\,\.\;\:\?\!\…])+",
    "stars": r"(\*)+",
}

text_groups = ["nl", "ws", "word"]
text_prog = re.compile("|".join(matchers[group] for group in text_groups))

markdown_groups = ["nl", "ws", "word", "stars"]
markdown_prog = re.compile("|".join(matchers[group] for group in markdown_groups))


def tokenize_text(text):
    """Splits the text in pieces of "nl", "ws" "word", and "other"."""
    return _tokenze(text, text_groups, text_prog)


def tokenize_markdown(text):
    """Splits the text in pieces of "nl", "ws" "word", "stars", and "other"."""
    return _tokenze(text, markdown_groups, markdown_prog)


def _tokenze(text, groups, prog):
    pos = 0
    while True:
        match = prog.search(text, pos)
        if match is None:
            break

        start = match.start()
        if start > pos:
            yield "other", text[pos : match.start()]

        yield groups[match.lastindex - 1], match.group()
        pos = match.end()

    other = text[pos:]
    if other:
        yield "other", other



import threading

import numpy as np

from ...resources import Texture, Buffer


def generate_size_table(max_size=8192):
    # 8192 is the default wgpu max texture size
    size = 0
    ref_area = 256
    while size < max_size:
        while size * size < ref_area:
            size += 8
        size = min(size, max_size)
        yield (size, size * size, ref_area)
        ref_area *= 2


# A table of predefined atlas sizes, resulting in ± factor-of-two increasing area.
# Each item in the table is a tuple (size, area, ref_area).
# Sizes are rounded to multiples of 8.
SIZES = list(generate_size_table())


def get_suitable_size(approximate_area):
    """Get the atlas size such that it has about the given area."""
    for i in range(1, len(SIZES)):
        size2, area2, _ = SIZES[i]
        if area2 >= approximate_area:
            size1, area1, _ = SIZES[i - 1]
            diff1 = np.abs(area1 - approximate_area)
            diff2 = np.abs(area2 - approximate_area)
            return size1 if diff1 < diff2 else size2
    else:
        return SIZES[-1][0]


class RectPacker:
    """
    This Python code was copied from the Vispy project.
    The algorithm is based on the article by Jukka Jylänki : "A Thousand Ways
    to Pack the Bin - A Practical Approach to Two-Dimensional Rectangle Bin
    Packing", February 27, 2010. More precisely, this is an implementation of
    the Skyline Bottom-Left algorithm based on C++ sources provided by Jukka
    Jylänki at: http://clb.demon.fi/files/RectangleBinPack/.
    """

    # Note: this could be extended to keep track of waste-space in a
    # separate data structure (a.k.a. waste map improvement). Similarly,
    # we could track freed regions for re-use. Right now we keep things
    # simple, and repack to re-claim freed regions.

    def _reset_packer(self):
        self._atlas_nodes = [(0, 0, self._array.shape[1])]

    def _select_region(self, width, height):
        """Find a free region of given size and allocate it.
        Returns the newly allocated region as (x, y, w, h), or None on failure.
        """

        # Find best place to fit it
        best_height = best_width = np.inf
        best_index = -1
        for i in range(len(self._atlas_nodes)):
            y = self._fit_region(i, width, height)
            if y >= 0:
                node = self._atlas_nodes[i]
                if y + height < best_height or (
                    y + height == best_height and node[2] < best_width
                ):
                    best_height = y + height
                    best_index = i
                    best_width = node[2]
                    region = node[0], y, width, height
        if best_index == -1:
            return None

        # Create the node and update the data structure
        node = region[0], region[1] + height, width
        self._atlas_nodes.insert(best_index, node)
        i = best_index + 1
        while i < len(self._atlas_nodes):
            node = self._atlas_nodes[i]
            prev_node = self._atlas_nodes[i - 1]
            if node[0] < prev_node[0] + prev_node[2]:
                shrink = prev_node[0] + prev_node[2] - node[0]
                x, y, w = self._atlas_nodes[i]
                self._atlas_nodes[i] = x + shrink, y, w - shrink
                if self._atlas_nodes[i][2] <= 0:
                    del self._atlas_nodes[i]
                    i -= 1
                else:
                    break
            else:
                break
            i += 1

        # Merge nodes
        i = 0
        while i < len(self._atlas_nodes) - 1:
            node = self._atlas_nodes[i]
            next_node = self._atlas_nodes[i + 1]
            if node[1] == next_node[1]:
                self._atlas_nodes[i] = node[0], node[1], node[2] + next_node[2]
                del self._atlas_nodes[i + 1]
            else:
                i += 1

        return region

    def _fit_region(self, index, width, height):
        """Test if region (width, height) fit into self._atlas_nodes[index].
        Returns y on success, or -1 on failure."""
        node = self._atlas_nodes[index]
        x, y = node[0], node[1]
        shape = self._array.shape
        width_left = width
        if x + width > shape[1]:
            return -1
        i = index
        while width_left > 0:
            node = self._atlas_nodes[i]
            y = max(y, node[1])
            if y + height > shape[0]:
                return -1
            width_left -= node[2]
            i += 1
        return y


class GlyphAtlas(RectPacker):
    """A global atlas for glyphs (thread-safe).

    This object allows allocating regions, which are packed into a large
    internal array. When the internal array is full, it is resized and
    repacked. This means that packed regions are moved to a new location
    (defragmentation). Therefore the actual location must be looked up
    via the infos array. The internal array is also reduced when the
    capacity is much larger than needed.
    """

    def __init__(self, initial_infos_size=1024, initial_array_size=1024):
        self._lock = threading.RLock()

        # Keep track of the index for each glyph
        self._hash2index = {}  # hash -> index
        self._index2hash = {}

        # Indices monotonically increase, but can also be reused from freed regions
        self._index_counter = 0
        self._free_indices = set()

        # Stats
        # The allocated_area represents the sum of areas that the packer has allocated,
        # not including the areas of freed regions. The free_area consists of regions
        # that were once allocated. The waste space in the atlas due to inefficient
        # packing is not counted.
        self._region_count = 0
        self._allocated_area = 0
        self._free_area = 0

        # Props to influence the behavior
        self.clear_free_regions = False
        self.downscale_ratio = 0.25

        # The per-glyph information (used in the shader)
        self._info_dtype = [
            ("origin", np.int32, 2),
            ("size", np.int32, 2),
            ("offset", np.float32, 2),
        ]

        # Init arrays
        self._infos = self._array = np.zeros((0,), np.void)
        self._initial_array_size = get_suitable_size(initial_array_size**2)
        self._set_new_infos_array(initial_infos_size)
        self._set_new_glyphs_array(self._initial_array_size)

    @property
    def region_count(self):
        """The number of regions allocated (excluding freed)."""
        return self._region_count

    @property
    def total_area(self):
        """The total area that the atlas could potentially hold."""
        shape = self._array.shape
        return shape[1] * shape[0]

    @property
    def allocated_area(self):
        """The allocated area (excluding the area of freed regions)."""
        return self._allocated_area

    def _set_new_infos_array(self, size):
        """Create a new array to store per-glyph info."""
        assert size > 0 and size > self._infos.shape[0]

        # Allocate new infos array
        infos1 = self._infos
        infos2 = np.zeros((size,), self._info_dtype)
        self._infos = infos2

        # Make a copy
        infos2[: infos1.shape[0]] = infos1

    def _set_new_glyphs_array(self, size):
        """Pack all glyphs up in a new array. If the size is unchanged,
        repacks the glyphs in the current array.
        """
        assert size > 0

        if size == self._array.shape[0]:
            # Keep the array, we'll repack only
            array1 = self._array.copy()
            array2 = self._array
            array2.fill(0)
        else:
            # Create new array
            array1 = self._array
            array2 = np.zeros((size, size), np.uint8)
            self._array = array2

        # We're going to pack it up fresh
        self._reset_packer()

        # Copy all rect regions to their new location
        allocated_area = 0
        for index in range(self._index_counter):
            info = self._infos[index]
            x1, y1 = info["origin"]
            w1, h1 = info["size"]
            # An index can have 0 size, for example the
            # character 32, a non-printable character.
            # has a glyph that is empty, but it should still
            # take space and have an index in our atlas.
            if w1 == 0 and index in self._free_indices:
                # freed region
                continue
            x2, y2, w2, h2 = self._select_region(w1, h1)
            info["origin"] = x2, y2
            array2[y2 : y2 + h2, x2 : x2 + w2] = array1[y1 : y1 + h1, x1 : x1 + w1]
            allocated_area += w2 * h2

        assert allocated_area == self._allocated_area

        self._free_area = 0
        self._allocated_area = allocated_area

    def allocate_region(self, w, h):
        """Allocate a region of the given size. Returns the index for
        the new region. The internal array may be repacked and/or
        resized if necessary.
        """
        with self._lock:
            # Select a region
            rect = self._select_region(w, h)

            # If the array is full the above returns None. If so, resize and try again.
            if rect is None:
                # If one third or more is empty, we repack only
                if self._free_area and self._free_area >= 0.5 * self._allocated_area:
                    self._set_new_glyphs_array(self._array.shape[0])
                    rect = self._select_region(w, h)
                # Increase size until we have enough space. Note that the requested
                # region might simply require more than double the original size.
                new_size = 0
                while rect is None:
                    new_size = get_suitable_size(self.total_area * 2)
                    if new_size**2 == self.total_area:
                        # You'd have to stretch things *a lot*, but this can happen in theory
                        raise RuntimeError(
                            "Glyph atlas is out of space and cannot be larger."
                        )
                    self._set_new_glyphs_array(new_size)
                    rect = self._select_region(w, h)

            # Select an index
            if self._free_indices:
                # Reuse an index that was previously freed.
                # Note that we only re-use the index (and thus the slot in the infos array),
                # but not the slot in the glyph array. Therefore _free_area does not change.
                index = self._free_indices.pop()
            else:
                # Create a new index
                index = self._index_counter
                self._index_counter += 1
                # Make sure that the per-glyph arrays are large enough
                while index >= self._infos.shape[0]:
                    self._set_new_infos_array(self._infos.shape[0] * 2)

            # Store the rectangle
            info = self._infos[index]
            info["origin"] = rect[:2]
            info["size"] = rect[2:]
            info["offset"] = 0, 0  # set to zero just in case

            # Bookkeeping
            self._region_count += 1
            self._allocated_area += w * h

            return index

    def set_region(self, index, region):
        """Set the region data for index."""
        with self._lock:
            info = self._infos[index]
            x, y = info["origin"]
            w, h = info["size"]
            self._array[y : y + h, x : x + w] = region

    def get_region(self, index):
        """Return a copy of the region corresponding to index."""
        with self._lock:
            info = self._infos[index]
            x, y = info["origin"]
            w, h = info["size"]
            return self._array[y : y + h, x : x + w].copy()

    def get_index_from_hash(self, hash):
        """Get the glyph index from the hash, or None on failure."""
        with self._lock:
            try:
                return self._hash2index[hash]
            except KeyError:
                return None

    def store_region_with_hash(self, hash, region, offset=(0, 0)):
        """Set the region corresponding to the given hash, and return the index."""
        with self._lock:
            # Sanitize
            assert isinstance(region, np.ndarray)
            assert hash not in self._hash2index

            # Reserve a slot
            w, h = region.shape[1], region.shape[0]
            index = self.allocate_region(w, h)

            # Store
            self.set_region(index, region)
            self._infos[index]["offset"] = offset

            # Bookkeeping
            self._hash2index[hash] = index
            self._index2hash[index] = hash

            return index

    def free_region(self, index):
        """Free up a region slot."""
        # We mask the index to only use the first 24 bits. In pygfx we use
        # the upper 8 bits to store font-prop info for the shader.
        index = int(index) & 0x0FFF
        with self._lock:
            # Zero out the region data
            if self.clear_free_regions:
                self.set_region(index, 0)
            # Free in data structure
            assert index < self._index_counter, "Invalid index to free"
            info = self._infos[index]
            w, h = info["size"]
            info["size"] = 0, 0
            self._free_indices.add(index)
            # Bookkeeping
            self._region_count -= 1
            self._free_area += w * h
            self._allocated_area -= w * h
            # Clear hash data
            hash = self._index2hash.pop(index, None)
            if hash is not None:
                self._hash2index.pop(hash)
            # Reduce size?
            if self.downscale_ratio:
                total_area = self._allocated_area + self._free_area
                if self._allocated_area <= self.downscale_ratio * total_area:
                    current_size = self._array.shape[0]
                    if current_size > self._initial_array_size:
                        new_size = get_suitable_size(self.total_area / 2)
                        new_size = max(new_size, self._initial_array_size)
                        if new_size != current_size:
                            self._set_new_glyphs_array(new_size)


class PygfxGlyphAtlas(GlyphAtlas):
    """A textured pygfx-specific subclass of the GlyphAtlas."""

    @property
    def texture(self):
        """The texture for the atlas. The Shared object exposes this object
        in a trackable way.
        """
        return self._texture

    @property
    def info_buffer(self):
        """A buffer containing the info for the glyphs (origin, size, offset)."""
        return self._infos_buffer

    def _set_new_glyphs_array(self, *args):
        # Do the normal behavior
        array = self._array
        super()._set_new_glyphs_array(*args)
        # Create resource object if needed
        if self._array is not array:
            self._texture = Texture(self._array, dim=2, force_contiguous=True)
        # Schedule an update. Note that the infos array is updated due to repacking.
        w, h = self._array.shape[1], self._array.shape[0]
        self._texture.update_range((0, 0, 0), (w, h, 1))
        self._infos_buffer.update_range(0, self._infos.shape[0])

    def _set_new_infos_array(self, *args):
        # Do the normal behavior
        infos = self._infos
        super()._set_new_infos_array(*args)
        # Create resource object if needed
        if self._infos is not infos:
            self._infos_buffer = Buffer(self._infos, force_contiguous=True)
        # Schedule an update
        self._infos_buffer.update_range(0, self._infos.shape[0])

    def set_region(self, index, glyph):
        with self._lock:
            super().set_region(index, glyph)
            info = self._infos[index]
            x, y = info["origin"]
            w, h = info["size"]
            self._texture.update_range((x, y, 0), (w, h, 1))
            self._infos_buffer.update_range(index, index + 1)


glyph_atlas = PygfxGlyphAtlas()



"""A very tiny CLI.

Invoke using e.g. ``python -m pygfx version``.
"""

import sys
import argparse

import pygfx


def main(argv=None):
    # Get argv so we can massage it
    if argv is None:
        argv = sys.argv
    if argv and argv[0].endswith(".py"):
        argv = argv[1:]

    # Defaults and aliases
    if not argv:
        argv = ["help"]
    if argv == ["--version"]:
        argv = ["version"]

    # Let the rest to argparse

    parser = argparse.ArgumentParser(
        prog="pygfx",
        description="The (very basic) Pygfx CLI",
    )

    parser.add_argument(
        "command", action="store", help="The command to run: 'help' or 'version'"
    )

    args = parser.parse_args(argv)
    command = args.command.lower()

    if command == "help":
        parser.print_help()
    if command == "version":
        print("pygfx v" + pygfx.__version__)
    else:
        print(f"Invalid command '{command}'")


if __name__ == "__main__":
    main()



import math

import numpy as np

from ..objects import Light

from .. import (
    sphere_geometry,
    Mesh,
    MeshBasicMaterial,
    Geometry,
    LineArrowMaterial,
    LineSegmentMaterial,
    Line,
)


class PointLightHelper(Mesh):
    """Light source indicator for point lights.

    Helper class to visualize a point light using some geometry (sphere by
    default). The helper object should be a parent of the light object.

    Parameters
    ----------
    size : float
        If geometry is None, the size of the indicator. Ignored otherwise.
    geometry : Geometry
        The geometry used to visualize the location of the light source. If None,
        a sphere will be used.
    color : Color
        The color of the sphere. If None, match the color of the light.

    """

    def __init__(self, size=1, geometry=None, color=None):
        assert isinstance(size, (int, float))
        if geometry is None:
            geometry = sphere_geometry(size)
        self._color = color
        material = MeshBasicMaterial(color="#fff")
        super().__init__(geometry, material)

    def _update_object(self):
        # Update on every draw
        super()._update_object()

        if self._color is None and isinstance(self.parent, Light):
            color = self.parent.color
            if color != self.material.color:
                self.material.color = color


class DirectionalLightHelper(Line):
    """Light source indicator for directional lights.

    Helper class to visualize a directional light. It shows arrows eminating
    from the light's position. If show_shadow_extent is True, it also shows the
    vector to the target and the extent of the shadowmap. The helper object
    should be a child of the light object.

    Parameters
    ----------
    ray_length : float
        The length of the indicator arrows.
    color : Color
        The color of the arrows. If None, match the color of the light.
    show_shadow_extent : bool
        If True, indicate the extent of the shadow map.

    """

    def __init__(self, ray_length=1, color=None, show_shadow_extent=False):
        self._color = color

        super().__init__(
            Geometry(positions=np.zeros((8, 3), np.float32)),
            LineArrowMaterial(color="#fff", thickness=5),
        )

        self._shadow_helper = Line(
            Geometry(positions=np.zeros((14, 3), np.float32)), LineSegmentMaterial()
        )
        self.add(self._shadow_helper)

        self.ray_length = ray_length
        self.show_shadow_extent = show_shadow_extent

    @property
    def ray_length(self):
        """The length of the arrows indicating light rays."""
        return self._ray_length

    @ray_length.setter
    def ray_length(self, value):
        self._ray_length = float(value)

        # Update geometry
        length = self._ray_length
        len5 = length / 5
        positions = np.array(
            [
                [len5, 0, 0],
                [len5, 0, -length],
                [-len5, 0, 0],
                [-len5, 0, -length],
                [0, len5, 0],
                [0, len5, -length],
                [0, -len5, 0],
                [0, -len5, -length],
            ],
            np.float32,
        )
        self.geometry.positions.data[:] = positions
        self.geometry.positions.update_range(0, 8)

    @property
    def show_shadow_extent(self):
        """Whether to also show the extent of the shadowmap."""
        return self._show_shadow_extent

    @show_shadow_extent.setter
    def show_shadow_extent(self, value):
        self._show_shadow_extent = bool(value)
        self._shadow_helper.visible = self._show_shadow_extent

    def _update_object(self):
        # Update on every draw.
        super()._update_object()

        if not isinstance(self.parent, Light):
            return

        if self._color is None:
            color = self.parent.color
            if color != self.material.color:
                self.material.color = color
                self._shadow_helper.material.color = color

        half_w = self.parent.shadow.camera.width / 2
        half_h = self.parent.shadow.camera.height / 2
        cur_size = np.abs(self._shadow_helper.geometry.positions.data[0])
        ref_size = (half_w, half_h, 0)

        if not np.isclose(cur_size, ref_size).all():
            positions = np.array(
                [
                    # Square
                    [-half_w, half_h, 0],
                    [half_w, half_h, 0],
                    [half_w, half_h, 0],
                    [half_w, -half_h, 0],
                    [half_w, -half_h, 0],
                    [-half_w, -half_h, 0],
                    [-half_w, -half_h, 0],
                    [-half_w, half_h, 0],
                    # Diagonals
                    [-half_w, -half_h, 0],
                    [half_w, half_h, 0],
                    [half_w, -half_h, 0],
                    [-half_w, half_h, 0],
                ],
                np.float32,
            )
            self._shadow_helper.geometry.positions.data[:12] = positions
            self._shadow_helper.geometry.positions.update_range(0, 12)

        lastval = -self.parent._gfx_distance_to_target
        if not np.isclose(lastval, self._shadow_helper.geometry.positions.data[13, 2]):
            self._shadow_helper.geometry.positions.data[13] = (0, 0, lastval)
            self._shadow_helper.geometry.positions.update_range(13, 14)


class SpotLightHelper(Line):
    """Light source indicator for spot lights.

    Helper class to visualize a spot light. The helper object should be a child
    of the light object.

    Parameters
    ----------
    color : Color
        The color of the arrows. If None, match the color of the light.

    """

    def __init__(self, color=None):
        self._color = color

        positions = [
            [0, 0, 0],
            [0, 0, -1],
            [0, 0, 0],
            [1, 0, -1],
            [0, 0, 0],
            [-1, 0, -1],
            [0, 0, 0],
            [0, 1, -1],
            [0, 0, 0],
            [0, -1, -1],
        ]

        for i in range(32):
            p1 = i / 32 * math.pi * 2
            p2 = (i + 1) / 32 * math.pi * 2

            positions.append([math.cos(p1), math.sin(p1), -1])
            positions.append([math.cos(p2), math.sin(p2), -1])

        super().__init__(
            Geometry(positions=positions),
            LineSegmentMaterial(thickness=1.0),
        )

    def _update_object(self):
        # Update on every draw
        super()._update_object()

        if not isinstance(self.parent, Light):
            return
        light = self.parent

        if self._color is None:
            color = light.color
            if color != self.material.color:
                self.material.color = color

        cone_length = light.distance or 1000
        cone_width = cone_length * math.tan(light.angle)
        self.local.scale = (cone_width, cone_width, cone_length)



import weakref


class WeakAssociativeContainer:
    """An object similar to a WeakKeyDictionary, except with multiple 'keys'."""

    def __init__(self):
        self.hash2value = {}  # hash -> value
        self.ref2hashes = {}  # ref -> hashes

        def remove(ref, selfref=weakref.ref(self)):  # noqa: B008
            self = selfref()
            if self is not None:
                hashes = self.ref2hashes.pop(ref, ())
                for h in hashes:
                    self.hash2value.pop(h, None)

        self._remove = remove

    def get(self, keys: tuple, default: object = None) -> object:
        """Get the value associated with a tuple of keys."""
        if not isinstance(keys, tuple):
            raise TypeError("WeakAssociativeContainer keys must be tuple.")
        return self.hash2value.get(hash(keys), default)

    def get_associated(self, key: object) -> set:
        """Get a set of values associated with the given (singular) key."""
        if isinstance(key, tuple):
            raise TypeError("WeakAssociativeContainer.get_all key must not be tuple.")
        hashes = self.ref2hashes.get(weakref.ref(key), ())
        values = []
        for h in hashes:
            try:
                values.append(self.hash2value[h])
            except KeyError:
                pass
        return set(values)

    def __getitem__(self, keys: tuple):
        if not isinstance(keys, tuple):
            raise TypeError("WeakAssociativeContainer keys must be tuple.")
        return self.hash2value[hash(keys)]

    def __setitem__(self, keys: tuple, value: object):
        if not isinstance(keys, tuple):
            raise TypeError("WeakAssociativeContainer keys must be tuple.")
        h = hash(keys)
        self.hash2value[h] = value
        for key in keys:
            ref = weakref.ref(key, self._remove)
            self.ref2hashes.setdefault(ref, set()).add(h)

    def setdefault(self, keys: tuple, default: object = None) -> object:
        try:
            return self[keys]
        except KeyError:
            self[keys] = default
            return default



"""
The purpose of a renderer is to render (i.e. draw) a scene to a canvas or
texture. It also provides picking, defines the anti-aliasing parameters, and any
post processing effects.

Classes
-------

.. autoclass:: pygfx.renderers.Renderer
    :members:

The system is designed such that different renderer implementations can
co-exist. However, :class:`~wgpu.WgpuRenderer` is likely the
only renderer you'll use.

The ``WgpuRenderer`` is capable of randering all types of objects. Other
renderers will likely be limited to a subset. E.g. the experimental
:class:`~svg.SVGRenderer` for lines, points and text.

Details
-------

A renderer is directly associated with its target and can only render to that
target. Different renderers can render to the same target though.

A renderer provides a ``.render()`` method that can be called one or more times
to render scenes. This creates a visual representation that is stored
internally, and is finally rendered into its render target (the canvas or
texture)::

                              __________
                             | blender  |
    [scenes] -- render() --> |  state   | -- flush() --> [target]
                             |__________|

The internal representation is managed by the blender object. The internal
render textures are typically at a higher resolution to reduce aliasing (SSAA).
The blender has auxiliary buffers such as a depth buffer, pick buffer, and
e.g. buffers for weighted fragments.

The flush-step resolves the internal representation into the target texture or
canvas, averaging neighbouring fragments for anti-aliasing.

"""

# ruff: noqa: F401, E402


class Renderer:
    """Base (abstract) renderer class that all renderers inherit from."""

    def render(self, scene, camera):
        """The method to call to render a scene from a camera viewpoint."""
        raise NotImplementedError()


from .wgpu import WgpuRenderer, print_wgpu_report
from .svg import SvgRenderer



from __future__ import annotations

import numpy as np
import pylinalg as la
from time import perf_counter_ns

from typing import Any, Tuple, Union


PRECISION_EPSILON = 1e-7


class cached:  # noqa: N801
    """Cache for computed properties.

    This descriptor implements a minimal timestamp-based cache for computed
    properties. The value of the property is computed using ``compute_fn`` and
    the result is cached until ``obj.last_modified`` advances. At this point the
    value of the computed property is recomputed upon the next read.

    """

    __slots__ = ("compute_fn", "name")

    def __init__(self, compute_fn=None) -> None:
        self.compute_fn = compute_fn
        self.name = None

    def __set_name__(self, clazz, name) -> None:
        self.name = f"cache_{name}_cache"

    def __get__(self, instance: AffineBase, clazz=None) -> Any:
        if instance is None:
            return self

        last_modified = instance.last_modified
        cache = getattr(instance, self.name, None)

        if cache is None or last_modified > cache[0]:
            cache = (last_modified, self.compute_fn(instance))
            setattr(instance, self.name, cache)

        return cache[1]


class AffineBase:
    """Base class for affine transformations.

    .. warning::
        In-place updates of slices of properties, e.g. ``transform.position[1] =
        42`` have no effect due to limitations of the python programming
        language and our decision to have the properties return pure numpy arrays.
        Where possible these arrays are flagged as read-only.

    This class implements basic getters and setters for the various properties
    of an affine transformation used in pygfx. If you are looking for
    `obj.local.<something>` or `obj.world.<something>` it is probably defined
    here.

    It also implements a basic caching mechanism that keeps computed properties
    around until the underlying transform changes. This makes use of the
    `@cached` descriptor defined above. The descriptor expects a `last_modified`
    property on the consuming class which is used as a monotonously increasing
    timestamp/counter to indicate if and when a cached value has become invalid.
    Once invalid, cached values are updated upon the next read/get meaning that
    they are updated lazily.

    Parameters
    ----------
    reference_up : ndarray, [3]
        The direction of the reference_up vector expressed in the target frame.
        It indicates neutral tilt and is used by the axis properties (right, up,
        forward) to maintain a common level of rotation around an axis when it
        is updated by it's setter. By default, it points along the Y-axis.
    is_camera_space : bool
        If True, the transform represents a camera space which means that it's
        ``forward`` and ``right`` directions are inverted.

    Notes
    -----
    Subclasses need to define and implement ``last_modified`` for the caching
    mechanism to work correctly. Check out existing subclasses for an example of
    how this might look like.

    All properties are **expressed in the target frame**, i.e., they use the
    target's basis, unless otherwise specified.

    """

    __slots__ = (
        "_reference_up",
        "_reference_up_view",
        "_scaling_signs",
        "_scaling_signs_view",
        "cache__decomposed_cache",
        "cache__decomposed_position_cache",
        "cache__directions_cache",
        "cache__euler_cache",
        "cache__inverse_matrix_cache",
        "cache__rotation_matrix_cache",
        "is_camera_space",
    )

    last_modified: int

    def __init__(self, /, *, reference_up=(0, 1, 0), is_camera_space=False):
        self.is_camera_space = int(is_camera_space)

        self._reference_up = la.vec_normalize(reference_up, dtype=float)
        self._reference_up_view = self._reference_up.view()
        self._reference_up_view.flags.writeable = False

        self._scaling_signs = np.asarray([1, 1, 1], dtype=float)
        self._scaling_signs_view = self._scaling_signs.view()
        self._scaling_signs_view.flags.writeable = False

    def flag_update(self):
        """Signal that this transform has updated."""
        raise NotImplementedError()

    @property
    def reference_up(self) -> np.ndarray:
        """The zero-tilt reference vector used for the direction setters."""
        return self._reference_up_view

    @reference_up.setter
    def reference_up(self, value):
        la.vec_normalize(value, out=self._reference_up)

    @property
    def matrix(self) -> np.ndarray:
        """Affine matrix describing this transform.

        ``vec_target = matrix @ vec_source``.

        """
        raise NotImplementedError()

    @cached
    def _inverse_matrix(self) -> np.ndarray:
        mat = la.mat_inverse(self.matrix)
        mat.flags.writeable = False
        return mat

    @property
    def scaling_signs(self) -> np.ndarray:
        """Property used to track and preserve the scale factor signs
        over matrix decomposition operations."""
        return self._scaling_signs_view

    @cached
    def _decomposed_position(self) -> np.ndarray:
        position = la.mat_decompose_translation(self.matrix)
        position.flags.writeable = False
        return position

    @cached
    def _decomposed(self) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        try:
            decomposed = la.mat_decompose(self.matrix, scaling_signs=self.scaling_signs)
        except ValueError:
            # the matrix has been set manually
            # and so there is no scaling component to preserve
            # any decomposed scaling is acceptable
            decomposed = la.mat_decompose(self.matrix)

        decomposed[0].flags.writeable = False
        decomposed[1].flags.writeable = False
        decomposed[2].flags.writeable = False

        return decomposed

    @cached
    def _directions(self):
        # Note: forward_is_minus_z indicates the camera frame
        directions = (2 * self.is_camera_space - 1, 1, 1 - 2 * self.is_camera_space)

        axes_source = np.diag(directions)
        axes_target = la.vec_transform(axes_source, self.matrix)
        origin_target = la.vec_transform((0, 0, 0), self.matrix)

        directions = axes_target - origin_target
        directions.flags.writeable = False
        return directions

    @cached
    def _rotation_matrix(self):
        rotation = la.mat_from_quat(self._decomposed[1])
        rotation.flags.writeable = False
        return rotation

    @cached
    def _euler(self):
        euler = la.quat_to_euler(self._decomposed[1])
        euler.flags.writeable = False
        return euler

    @property
    def inverse_matrix(self) -> np.ndarray:
        """Inverse of this transform as affine matrix.

        ``vec_source = inverse_matrix @ vec_target``

        """
        return self._inverse_matrix

    @property
    def position(self) -> np.ndarray:
        """The origin of source."""
        return self._decomposed_position

    @property
    def rotation(self) -> np.ndarray:
        """The orientation of source as a quaternion."""
        return self._decomposed[1]

    @property
    def rotation_matrix(self) -> np.ndarray:
        """The orientation of source as a rotation matrix."""
        return self._rotation_matrix

    @property
    def euler(self) -> np.ndarray:
        """The orientation of source as XYZ euler angles."""
        return self._euler

    @property
    def euler_x(self) -> float:
        """The X component of source's orientation as XYZ euler angles."""
        return self._euler[0]

    @property
    def euler_y(self) -> float:
        """The Y component of source's orientation as XYZ euler angles."""
        return self._euler[1]

    @property
    def euler_z(self) -> float:
        """The Z component of source's orientation as XYZ euler angles."""
        return self._euler[2]

    @property
    def scale(self) -> np.ndarray:
        """The scale of source."""
        return self._decomposed[2]

    @property
    def x(self) -> float:
        """The X component of source's position."""
        return self.position[0]

    @property
    def y(self) -> float:
        """The Y component of source's position."""
        return self.position[1]

    @property
    def z(self) -> float:
        """The Z component of source's position."""
        return self.position[2]

    @property
    def scale_x(self) -> float:
        """The X component of source's scale."""
        return self.scale[0]

    @property
    def scale_y(self) -> float:
        """The Y component of source's scale."""
        return self.scale[1]

    @property
    def scale_z(self) -> float:
        """The Z component of source's scale."""
        return self.scale[2]

    @property
    def right(self) -> np.ndarray:
        """The right direction of source."""
        return self._directions[0]

    @property
    def up(self) -> np.ndarray:
        """The up direction of source."""
        return self._directions[1]

    @property
    def forward(self) -> np.ndarray:
        """The forward direction of source."""
        return self._directions[2]

    @position.setter
    def position(self, value):
        self.matrix = la.mat_compose(value, self.rotation, self.scale)

    @rotation.setter
    def rotation(self, value):
        self.matrix = la.mat_compose(self.position, value, self.scale)

    @euler.setter
    def euler(self, value):
        self.rotation = la.quat_from_euler(value)

    @rotation_matrix.setter
    def rotation_matrix(self, value):
        self.rotation = la.quat_from_mat(value)

    @euler_x.setter
    def euler_x(self, value):
        self.euler = (value, 0, 0)

    @euler_y.setter
    def euler_y(self, value):
        self.euler = (0, value, 0)

    @euler_z.setter
    def euler_z(self, value):
        self.euler = (0, 0, value)

    @scale.setter
    def scale(self, value):
        m = la.mat_compose(self.position, self.rotation, value)
        np.sign(value, out=self._scaling_signs)
        self.matrix = m

    @property
    def components(self) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        return self._decomposed

    @components.setter
    def components(self, value: Tuple[np.ndarray, np.ndarray, np.ndarray]):
        m = la.mat_compose(*value)
        np.sign(value[2], out=self._scaling_signs)
        self.matrix = m

    @x.setter
    def x(self, value):
        _, y, z = self.position
        self.position = (value, y, z)

    @y.setter
    def y(self, value):
        x, _, z = self.position
        self.position = (x, value, z)

    @z.setter
    def z(self, value):
        x, y, _ = self.position
        self.position = (x, y, value)

    @scale_x.setter
    def scale_x(self, value):
        _, y, z = self.scale
        self.scale = (value, y, z)

    @scale_y.setter
    def scale_y(self, value):
        x, _, z = self.scale
        self.scale = (x, value, z)

    @scale_z.setter
    def scale_z(self, value):
        x, y, _ = self.scale
        self.scale = (x, y, value)

    @right.setter
    def right(self, value):
        value = la.vec_normalize(value)

        if np.allclose(value, (0, 0, 0)):
            raise ValueError("A coordinate axis can't point at its origin.")

        if np.linalg.norm(np.cross(value, self.reference_up)) == 0:
            # target and reference_up are parallel
            rotation = la.quat_from_vecs(self.forward, value)
        else:
            matrix = la.mat_look_at((0, 0, 0), value, self.reference_up)
            rotation = la.quat_from_mat(matrix)

        if self.is_camera_space:
            part2 = la.quat_from_axis_angle((0, 1, 0), -np.pi / 2)
        else:
            part2 = la.quat_from_axis_angle((0, 1, 0), np.pi / 2)

        self.rotation = la.quat_mul(rotation, part2)

    @up.setter
    def up(self, value):
        value = la.vec_normalize(value)

        if np.allclose(value, (0, 0, 0)):
            raise ValueError("A coordinate axis can't point at its origin.")

        if np.linalg.norm(np.cross(value, self.reference_up)) == 0:
            # target and reference_up are parallel
            rotation = la.quat_from_vecs(self.forward, value)
        else:
            matrix = la.mat_look_at((0, 0, 0), value, self.reference_up)
            rotation = la.quat_from_mat(matrix)

        part2 = la.quat_from_axis_angle((1, 0, 0), np.pi / 2)
        self.rotation = la.quat_mul(rotation, part2)

    @forward.setter
    def forward(self, value):
        value = la.vec_normalize(value)

        if np.allclose(value, (0, 0, 0)):
            raise ValueError("A coordinate axis can't point at its origin.")

        if np.linalg.norm(np.cross(value, self.reference_up)) < PRECISION_EPSILON:
            # target and reference_up are parallel
            if np.linalg.norm(np.cross(self.forward, value)) < PRECISION_EPSILON:
                # target and forward are parallel, no rotation needed
                return
            rotation = la.quat_from_vecs(self.forward, value)
        elif self.is_camera_space:
            matrix = la.mat_look_at((0, 0, 0), -value, self.reference_up)
            rotation = la.quat_from_mat(matrix)
        else:
            matrix = la.mat_look_at((0, 0, 0), value, self.reference_up)
            rotation = la.quat_from_mat(matrix)

        self.rotation = rotation

    def __array__(self, dtype=None) -> np.ndarray:
        return self.matrix.astype(dtype, copy=False)


class AffineTransform(AffineBase):
    """A single affine transform.

    Parameters
    ----------
    position : ndarray, [3]
        The position of this transform expressed in the target frame.
        This argument is only consumed if ``state_basis`` is set to "components".
    rotation : ndarray, [4]
        The rotation quaternion of this transform expressed in the target frame.
        This argument is only consumed if ``state_basis`` is set to "components".
    scale : ndarray, [3]
        The per-axis scale of this transform expressed in the target frame.
        This argument is only consumed if ``state_basis`` is set to "components".
    reference_up : ndarray, [3]
        The direction of the reference_up vector expressed in the target frame.
        It indicates neutral tilt and is used by the axis properties (right, up,
        forward) to maintain a common level of rotation around an axis when it
        is updated by it's setter. By default, it points along the Y-axis.
    is_camera_space : bool
        If True, the transform represents a camera space which means that it's
        ``forward`` and ``right`` directions are inverted.
    state_basis : str
        One of "components" or "matrix". The default is "components".
        If "components", the transform will store its state in the form of
        position, rotation, and scale, while the matrix will be computed.
        If "matrix", the transform will store its state in the form of a matrix,
        while the position, rotation and scale will be computed.
        This option is provided to allow for performance optimizations when
        the transform is used in a context where the state is updated in one
        form but queried in another. Depending on your usage patterns, you
        may avoid significant overhead by choosing the appropriate basis.
    matrix : ndarray, [4, 4]
        The affine matrix describing this transform.
        This argument is only consumed if ``state_basis`` is set to "matrix".

    Notes
    -----
    When updating the underlying numpy arrays in-place these updates will not
    propagate via the transform's callback system nor will they invalidate
    existing caches. To inform the transform of these updates call
    ``transform.flag_update()``, which will trigger both callbacks and cache
    invalidation.

    See Also
    --------
    AffineBase
        Base class defining various useful properties for this transform.

    """

    __slots__ = (
        "_matrix",
        "_matrix_view",
        "_position",
        "_position_view",
        "_rotation",
        "_rotation_view",
        "_scale",
        "_scale_view",
        "_state_basis",
        "_wrapper",
        "cache__composed_matrix_cache",
        "cache__computed_scaling_signs_cache",
        "last_modified",
    )

    def __init__(
        self,
        position=(0, 0, 0),
        rotation=(0, 0, 0, 1),
        scale=(1, 1, 1),
        *,
        reference_up=(0, 1, 0),
        is_camera_space=False,
        state_basis="components",
        matrix: np.ndarray = None,
    ) -> None:
        super().__init__(reference_up=reference_up, is_camera_space=is_camera_space)
        self._state_basis = state_basis
        self.last_modified = perf_counter_ns()

        self._position = np.asarray(position, dtype=float)
        self._rotation = np.asarray(rotation, dtype=float)
        self._scale = np.asarray(scale, dtype=float)

        self._position_view = self._position.view()
        self._position_view.flags.writeable = False
        self._rotation_view = self._rotation.view()
        self._rotation_view.flags.writeable = False
        self._scale_view = self._scale.view()
        self._scale_view.flags.writeable = False

        # The ._matrix is only used when state_basis is "matrix"
        if state_basis == "matrix" and matrix is not None:
            self._matrix = np.asarray(matrix, dtype=float)
        else:
            self._matrix = np.identity(4, dtype=float)

        self._matrix_view = self._matrix.view()
        self._matrix_view.flags.writeable = False

        self._wrapper: RecursiveTransform = None

    def flag_update(self):
        """Signal that this transform has updated."""
        # note: this function has been heavily micro-optimized
        # please don't modify it carelessly
        last_modified = perf_counter_ns()
        self.last_modified = last_modified
        if wrapper := self._wrapper:
            wrapper.flag_update(last_modified)

    def _set_wrapper(self, wrapper: RecursiveTransform):
        self._wrapper = wrapper

    @property
    def reference_up(self) -> np.ndarray:
        """The zero-tilt reference vector used for the direction setters."""
        if self._wrapper:
            return self._wrapper._parent_reference_up
        return super().reference_up

    @reference_up.setter
    def reference_up(self, value):
        if self._wrapper:
            self._wrapper._parent_reference_up = value
            return
        AffineBase.reference_up.fset(self, value)

    @property
    def state_basis(self) -> str:
        """The basis of the transform, either "components" (default) or "matrix"."""
        return self._state_basis

    @state_basis.setter
    def state_basis(self, value: str):
        if value not in ("components", "matrix"):
            raise ValueError("state_basis must be either 'components' or 'matrix'")
        if value == "matrix":
            self._matrix[:] = self.matrix
        elif value == "components":
            self._position[:] = self.position
            self._rotation[:] = self.rotation
            self._scale[:] = self.scale
        self._state_basis = value
        self.flag_update()

    @property
    def components(self) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        if self.state_basis == "components":
            return (self._position_view, self._rotation_view, self._scale_view)
        return super().components

    @components.setter
    def components(self, value: Tuple[np.ndarray, np.ndarray, np.ndarray]):
        if self.state_basis == "components":
            if (pos := value[0]) is not None:
                self._position[:] = pos
            if (rot := value[1]) is not None:
                self._rotation[:] = rot
            if (sca := value[2]) is not None:
                self._scale[:] = sca
        else:
            la.mat_compose(*value, out=self._matrix)
            np.sign(value[2], out=self._scaling_signs)
        self.flag_update()

    @property
    def position(self) -> np.ndarray:
        """The origin of source."""
        if self.state_basis == "components":
            return self._position_view
        return super().position

    @position.setter
    def position(self, value):
        if self.state_basis == "components":
            self._position[:] = value
        else:
            la.mat_compose(value, self.rotation, self.scale, out=self._matrix)
        self.flag_update()

    @property
    def rotation(self) -> np.ndarray:
        """The orientation of source as a quaternion."""
        if self.state_basis == "components":
            return self._rotation_view
        return super().rotation

    @rotation.setter
    def rotation(self, value):
        if self.state_basis == "components":
            self._rotation[:] = value
        else:
            la.mat_compose(self.position, value, self.scale, out=self._matrix)
        self.flag_update()

    @property
    def scale(self) -> np.ndarray:
        """The scale of source."""
        if self.state_basis == "components":
            return self._scale_view
        return super().scale

    @scale.setter
    def scale(self, value):
        if self.state_basis == "components":
            self._scale[:] = value
        else:
            la.mat_compose(self.position, self.rotation, value, out=self._matrix)
            np.sign(value, out=self._scaling_signs)
        self.flag_update()

    @cached
    def _computed_scaling_signs(self) -> np.ndarray:
        signs = np.sign(self._scale)
        signs.flags.writeable = False
        return signs

    @property
    def scaling_signs(self) -> np.ndarray:
        """Property used to track and preserve the scale factor signs
        over matrix decomposition operations."""
        if self.state_basis == "components":
            return self._computed_scaling_signs
        return super().scaling_signs

    @cached
    def _rotation_matrix(self) -> np.ndarray:
        rotation = la.mat_from_quat(self._rotation)
        rotation.flags.writeable = False
        return rotation

    @property
    def rotation_matrix(self) -> np.ndarray:
        """The orientation of source as a rotation matrix."""
        if self.state_basis == "components":
            return self._rotation_matrix
        return super().rotation_matrix

    rotation_matrix = rotation_matrix.setter(AffineBase.rotation_matrix.fset)

    @cached
    def _euler(self) -> np.ndarray:
        euler = la.quat_to_euler(self._rotation)
        euler.flags.writeable = False
        return euler

    @property
    def euler(self) -> np.ndarray:
        """The orientation of source as XYZ euler angles."""
        if self.state_basis == "components":
            return self._euler
        return super().euler

    euler = euler.setter(AffineBase.euler.fset)

    @cached
    def _composed_matrix(self) -> np.ndarray:
        mat = la.mat_compose(self._position, self._rotation, self._scale)
        mat.flags.writeable = False
        return mat

    @property
    def matrix(self) -> np.ndarray:
        """Affine matrix describing this transform.

        ``vec_target = matrix @ vec_source``.

        """
        if self.state_basis == "components":
            return self._composed_matrix
        return self._matrix_view

    @matrix.setter
    def matrix(self, value: np.ndarray):
        if self.state_basis == "components":
            try:
                # when the matrix is set manually
                # try to maintain the most recent configured scaling signs
                position, rotation, scale = la.mat_decompose(
                    value, scaling_signs=self.scaling_signs
                )
            except ValueError:
                position, rotation, scale = la.mat_decompose(value)
            self._position[:] = position
            self._rotation[:] = rotation
            self._scale[:] = scale
        elif self.state_basis == "matrix":
            self._matrix[:] = value
        self.flag_update()

    def __matmul__(self, other) -> Union[AffineTransform, np.ndarray]:
        if isinstance(other, AffineTransform):
            matrix = self.matrix @ other.matrix

            state_basis = self.state_basis
            if la.mat_has_shear(matrix):
                # if the resulting transform has shearing
                # force matrix state_basis - we don't
                # support shearing in components, see #920
                state_basis = "matrix"

            kwargs = dict(
                is_camera_space=self.is_camera_space,
                state_basis=state_basis,
                matrix=matrix,
            )
            if state_basis == "components":
                try:
                    decomposed = la.mat_decompose(
                        matrix,
                        scaling_signs=self.scaling_signs * other.scaling_signs,
                    )
                except ValueError:
                    decomposed = la.mat_decompose(matrix)
                kwargs["position"] = decomposed[0]
                kwargs["rotation"] = decomposed[1]
                kwargs["scale"] = decomposed[2]

            return AffineTransform(**kwargs)

        return np.asarray(self) @ other


class RecursiveTransform(AffineBase):
    """A transform that may be preceded by another transform.

    This transform behaves semantically identical to an ordinary
    ``AffineTransform`` (same properties), except that users may define a
    ``parent`` transform which precedes the ``matrix`` used by the ordinary
    ``AffineTransform``. The resulting ``RecursiveTransform`` then controls the
    total transform that results from combining the two transforms via::

        recursive_transform = parent @ matrix

    In other words, the source frame of ``RecursiveTransform`` is the source
    frame of ``matrix`` and the target frame of ``RecursiveTransform`` is the
    target frame of ``parent``. Implying that ``RecursiveTransform``'s
    properties are given in the target frame.

    The use case for this class is to allow getting and setting of properties of
    a WorldObjects world transform, i.e., it implements ``WorldObject.world``.

    Under the hood, this transform wraps another transform (passed in as
    ``base``), similar to how ``AffineTransform`` wraps a numpy array.
    Setting properties of ``RecursiveTransform`` will internally
    transform the new values into ``matrix`` target frame and then set the
    obtained value on the wrapped transform. This means that the ``parent``
    transform is not affected by changes made to this transform.

    Parameters
    ----------
    base : AffineTransform
        The base transform that will be wrapped by this transform.
    parent : RecursiveTransform, optional
        The parent transform that precedes the base transform.
    reference_up : ndarray, [3]
        The direction of the reference_up vector expressed in the target
        frame. It is used by the axis properties (right, up, forward)
        to maintain a common level of rotation around an axis when it
        is updated by it's setter. By default, it points along the
        positive Y-axis.
    is_camera_space : bool
        If True, the transform represents a camera space which means that it's
        ``forward`` and ``right`` directions are inverted.

    See Also
    --------
    AffineBase
        Base class defining various useful properties for this transform.

    """

    __slots__ = (
        "_parent",
        "cache__computed_parent_reference_up_cache",
        "cache__computed_scaling_signs_cache",
        "cache__matrix_cache",
        "children",
        "last_modified",
        "own",
    )

    def __init__(
        self,
        base: AffineTransform,
        /,
        *,
        parent: RecursiveTransform = None,
        reference_up=(0, 1, 0),
        is_camera_space=False,
    ) -> None:
        super().__init__(reference_up=reference_up, is_camera_space=is_camera_space)
        self.last_modified = perf_counter_ns()
        self._parent = parent
        self.children = []
        self.own = base
        self.own._set_wrapper(self)

    def flag_update(self, last_modified=None):
        """Signal that this transform has updated."""
        # note: this function has been heavily micro-optimized
        # please don't modify it carelessly
        if last_modified is None:
            last_modified = perf_counter_ns()
        self.last_modified = last_modified
        if children := self.children:
            for child in children:
                if child.children:
                    child.flag_update(last_modified)
                else:
                    child.last_modified = last_modified

    @cached
    def _computed_parent_reference_up(self) -> np.ndarray:
        """The direction of the reference_up vector expressed in the parent frame."""
        new_ref = la.vec_transform(self._reference_up, self._parent.inverse_matrix)
        origin = la.vec_transform((0, 0, 0), self._parent.inverse_matrix)
        vec = la.vec_normalize(new_ref - origin)
        vec.flags.writeable = False
        return vec

    @property
    def _parent_reference_up(self) -> np.ndarray:
        if self._parent:
            return self._computed_parent_reference_up
        return self._reference_up_view

    @_parent_reference_up.setter
    def _parent_reference_up(self, value):
        if self._parent:
            new_ref = la.vec_transform(value, self._parent.matrix)
            origin = self._parent.position
            la.vec_normalize(new_ref - origin, out=self._reference_up)
        else:
            la.vec_normalize(value, out=self._reference_up)
        self.flag_update()
        # Note: since _parent_reference_up is only used in a setter in AffineBase
        # we do not need to call flag_update() on self.own; all its state and cache
        # remains intact

    @property
    def parent(self) -> RecursiveTransform:
        """The transform that preceeds the own/local transform."""
        return self._parent

    @parent.setter
    def parent(self, value: RecursiveTransform):
        self._parent = value
        self.flag_update()

    @cached
    def _matrix(self) -> np.ndarray:
        mat = self._parent.matrix @ self.own.matrix
        mat.flags.writeable = False
        return mat

    @property
    def matrix(self) -> np.ndarray:
        """Affine matrix describing this transform.

        ``vec_target = matrix @ vec_source``.

        """
        if self._parent:
            return self._matrix
        return self.own.matrix

    @matrix.setter
    def matrix(self, value: np.ndarray):
        if self._parent:
            if self.own.state_basis == "matrix":
                np.matmul(self._parent.inverse_matrix, value, out=self.own._matrix)
                self.own.flag_update()
            else:
                self.own.matrix = self._parent.inverse_matrix @ value
        else:
            self.own.matrix = value

    def __matmul__(self, other) -> Union[RecursiveTransform, np.ndarray]:
        if isinstance(other, AffineBase):
            return RecursiveTransform(other, parent=self)
        else:
            return np.asarray(self) @ other

    @cached
    def _computed_scaling_signs(self) -> np.ndarray:
        signs = self._parent.scaling_signs * self.own.scaling_signs
        signs.flags.writeable = False
        return signs

    @property
    def scaling_signs(self) -> np.ndarray:
        """Property used to track and preserve the scale factor signs
        over matrix decomposition operations."""
        if self._parent:
            return self._computed_scaling_signs
        return self.own.scaling_signs



import wgpu

from ....resources import Buffer
from ....utils import logger


from .utils import to_texture_format, GfxSampler, GfxTextureView
from .update import ensure_wgpu_object, ALTTEXFORMAT


visibility_render = wgpu.ShaderStage.VERTEX | wgpu.ShaderStage.FRAGMENT
visibility_all = (
    wgpu.ShaderStage.VERTEX | wgpu.ShaderStage.FRAGMENT | wgpu.ShaderStage.COMPUTE
)


class Binding:
    """Simple object to hold together some information about a binding, for internal use.

    Parameters:
        name: the name in wgsl
        type: "buffer/subtype", "sampler/subtype", "texture/subtype", "storage_texture/subtype".
            The subtype depends on the type:
            BufferBindingType, SamplerBindingType, TextureSampleType, or StorageTextureAccess.
        resource: the Buffer, GfxTextureView, or GfxSampler object.
        visibility: wgpu.ShaderStage flag
        structname: the custom wgsl struct name, if any. otherwise will auto-generate.

    Some tips on terminology:

    * A "binding" is a generic term for an object that defines how a
        resource (buffer or texture) is bound to a shader. In this subpackage it
        likely means a Binding object like this.
    * This binding can be represented with a binding_descriptor and
        binding_layout_descriptor. These are dicts to be passed to wgpu.
    * A list of these binding_layout_descriptor's can be passed to create_bind_group_layout.
    * A list of these binding_layout's can be passed to create_bind_group.
    * Multiple bind_group_layout's can be combined into a pipeline_layout.

    """

    def __init__(
        self, name, type, resource, visibility=visibility_render, structname=None
    ):
        if isinstance(visibility, str):
            visibility = getattr(wgpu.ShaderStage, visibility)
        self.name = name
        self.type = type
        self.resource = resource
        self.visibility = visibility

        self.structname = structname

    def _require_usage_flags(self, resource, usage_flags):
        if resource._wgpu_object is None:
            resource._wgpu_usage |= usage_flags
        elif not (resource._wgpu_usage & usage_flags):
            logger.warning(
                "{resource} requires usage {usage_flags}, but has already been created."
            )

    def get_bind_group_descriptors(self, slot):
        """Get the descriptors (dicts) for creating a binding_descriptor
        and binding_layout_descriptor. A list of these descriptors are
        combined into a bind_group and bind_group_layout.
        """
        resource = self.resource
        subtype = self.type.partition("/")[2]

        if self.type.startswith("buffer/"):
            assert isinstance(resource, Buffer)
            usage_flags = 0
            if "uniform" in self.type:
                usage_flags |= wgpu.BufferUsage.UNIFORM
            elif "storage" in self.type:
                usage_flags |= wgpu.BufferUsage.STORAGE
                if "indices" in self.name:
                    usage_flags |= wgpu.BufferUsage.INDEX
                else:
                    usage_flags |= wgpu.BufferUsage.VERTEX
            self._require_usage_flags(resource, usage_flags)
            binding = {
                "binding": slot,
                "resource": {
                    "buffer": ensure_wgpu_object(resource),
                    "offset": 0,
                    "size": resource.nbytes,
                },
            }
            # Determine min_binding_size. We can set it to None, so it won't do
            # an early check, causing runtime checks to be needed. If we want to
            # specify it, the needed value is different for shader array
            # variables that have a static length vs ones that have a length
            # defined at runtime. Then there's caching, which isn't really used
            # well when min_binding_size jumps around. On the other hand, in early
            # versions we had a comment here that suggested that the caching
            # caused re-use of binding layouts, but them not being compatible.
            # So let's try this: set the min_binding_size for uniform arrays,
            # and set to default for storage buffers.
            # Also see https://github.com/pygfx/pygfx/issues/855
            if "uniform" in self.type:
                min_binding_size = resource.nbytes
            else:
                min_binding_size = None  # None or resource.itemsize
            binding_layout = {
                "binding": slot,
                "visibility": self.visibility,
                "buffer": {
                    "type": getattr(wgpu.BufferBindingType, subtype),
                    "has_dynamic_offset": False,
                    "min_binding_size": min_binding_size,
                },
            }
        elif self.type.startswith("sampler/"):
            assert isinstance(resource, GfxSampler)
            binding = {
                "binding": slot,
                "resource": ensure_wgpu_object(resource),
            }
            binding_layout = {
                "binding": slot,
                "visibility": self.visibility,
                "sampler": {
                    "type": getattr(wgpu.SamplerBindingType, subtype),
                },
            }
        elif self.type.startswith("texture/"):
            assert isinstance(resource, GfxTextureView)
            self._require_usage_flags(
                resource.texture, wgpu.TextureUsage.TEXTURE_BINDING
            )
            binding = {
                "binding": slot,
                "resource": ensure_wgpu_object(resource),
            }
            dim = resource.view_dim
            dim = getattr(wgpu.TextureViewDimension, dim, dim)
            sample_type = getattr(wgpu.TextureSampleType, subtype, subtype)
            # Derive sample type from texture
            if sample_type == "auto":
                fmt = to_texture_format(resource.format)
                fmt = ALTTEXFORMAT.get(fmt, [fmt])[0]
                if "float" in fmt or "norm" in fmt:
                    sample_type = wgpu.TextureSampleType.float
                    # For float32 wgpu does not allow the sampler to be filterable,
                    # except when the native-only feature
                    # TEXTURE_ADAPTER_SPECIFIC_FORMAT_FEATURES is set,
                    # which wgpu-py does by default.
                    # if "32float" in fmt:
                    #     sample_type = wgpu.TextureSampleType.unfilterable_float
                elif "uint" in fmt:
                    sample_type = wgpu.TextureSampleType.uint
                elif "sint" in fmt:
                    sample_type = wgpu.TextureSampleType.sint
                elif "depth" in fmt:
                    sample_type = wgpu.TextureSampleType.depth
                else:
                    raise ValueError("Could not determine texture sample type.")
            binding_layout = {
                "binding": slot,
                "visibility": self.visibility,
                "texture": {
                    "sample_type": sample_type,
                    "view_dimension": dim,
                    "multisampled": False,
                },
            }
        elif self.type.startswith("storage_texture/"):
            assert isinstance(resource, GfxTextureView)
            self._require_usage_flags(
                resource.texture, wgpu.TextureUsage.STORAGE_BINDING
            )
            binding = {
                "binding": slot,
                "resource": ensure_wgpu_object(resource),
            }
            dim = resource.view_dim
            dim = getattr(wgpu.TextureViewDimension, dim, dim)
            fmt = to_texture_format(resource.format)
            fmt = ALTTEXFORMAT.get(fmt, [fmt])[0]
            binding_layout = {
                "binding": slot,
                "visibility": self.visibility,
                "storage_texture": {
                    "access": getattr(wgpu.StorageTextureAccess, subtype),
                    "format": fmt,
                    "view_dimension": dim,
                },
            }
        elif self.type.startswith("shadow_texture/"):
            # a shadow_texture's resource is wgpu.GPUTextureView
            assert isinstance(resource, wgpu.GPUTextureView)
            binding = {"binding": slot, "resource": resource}

            binding_layout = {
                "binding": slot,
                "visibility": wgpu.ShaderStage.FRAGMENT,
                "texture": {
                    "sample_type": wgpu.TextureSampleType.depth,
                    "view_dimension": subtype,
                    "multisampled": False,
                },
            }
        elif self.type.startswith("shadow_sampler/"):
            # a shadow_sampler's resource is wgpu.GPUSampler
            assert isinstance(resource, wgpu.GPUSampler)
            binding = {"binding": slot, "resource": resource}

            binding_layout = {
                "binding": slot,
                "visibility": wgpu.ShaderStage.FRAGMENT,
                "sampler": {"type": wgpu.SamplerBindingType.comparison},
            }
        else:
            raise RuntimeError(f"Unexpected binding type: '{self.type}'")

        return binding, binding_layout



import math
from .interpolant import LinearInterpolant
from ..objects import Event


class AnimationAction:
    """An AnimationAction represents a specific animation that can be played within the AnimationMixer.

    It is a single animation clip that can be played, paused, stopped, faded in/out, and cross-faded with other actions.
    """

    def __init__(self, mixer, clip):
        self._mixer = mixer
        self._clip = clip

        self.weight = 1.0
        self._effective_weight = 1.0

        self.time = 0
        self.time_scale = 1.0
        self._effective_time_scale = 1.0

        self.repetitions = math.inf

        self.paused = False
        self.enabled = True

        self._weight_interpolant = None
        self._time_scale_interpolant = None

    def play(self):
        """Play the action."""
        self._mixer._activate_action(self)
        return self

    def stop(self):
        """Stop the action."""
        self._mixer._deactivate_action(self)
        return self.reset()

    def reset(self):
        """Reset the action."""
        self.paused = False
        self.enabled = True

        self.time = 0

        return self.stop_fading()

    def is_running(self):
        return (
            self.enabled
            and not self.paused
            and self.time_scale != 0
            and self._mixer._is_active_action(self)
        )

    def is_schedule(self):
        # return True when play is called
        return self._mixer._is_active_action(self)

    def set_loop(self, repetitions=math.inf):
        """Set the number of repetitions for the action."""
        self.repetitions = repetitions
        return self

    @property
    def effective_weight(self):
        return self._effective_weight

    def set_effective_weight(self, weight):
        """Set the effective weight for the action."""
        self.weight = weight
        self._effective_weight = weight if self.enabled else 0

        return self.stop_fading()

    def fade_in(self, duration):
        """Fade in the action."""
        return self._schedule_fading(duration, 0, 1)

    def fade_out(self, duration):
        """Fade out the action."""
        return self._schedule_fading(duration, 1, 0)

    def cross_fade_from(self, fade_out_action: "AnimationAction", duration, warp):
        """Cross fade from another action."""
        fade_out_action.fade_out(duration)
        self.fade_in(duration)

        if warp:
            fade_in_duration = self._clip.duration
            fade_out_duration = fade_out_action._clip.duration

            start_end_ratio = fade_out_duration / fade_in_duration
            end_start_ratio = fade_in_duration / fade_out_duration

            fade_out_action.warp(1.0, start_end_ratio, duration)
            self.warp(end_start_ratio, 1.0, duration)
        return self

    def cross_fade_to(self, fade_in_action: "AnimationAction", duration, warp):
        """Cross fade to another action."""
        return fade_in_action.cross_fade_from(self, duration, warp)

    def stop_fading(self):
        """Stop the fading of the action."""
        self._weight_interpolant = None
        return self

    @property
    def effective_time_scale(self):
        return self._effective_time_scale

    def set_effective_time_scale(self, time_scale):
        """Set the effective time scale for the action."""
        self.time_scale = time_scale
        self._effective_time_scale = 0 if self.paused else time_scale
        return self.stop_warping()

    def set_duration(self, duration):
        """Set the duration of the action by adjusting the time scale."""
        self.time_scale = self._clip.duration / duration
        return self.stop_warping()

    def halt(self, duration):
        """Halt the action."""
        return self.warp(self._effective_time_scale, 0, duration)

    def warp(self, start_time_scale, end_time_scale, duration):
        """Warp the action."""
        now = self._mixer._time
        interpolant = self._time_scale_interpolant
        time_scale = self.time_scale

        if interpolant is None:
            times = [now, now + duration]
            values = [start_time_scale, end_time_scale]
            interpolant = LinearInterpolant(times, values)
            self._time_scale_interpolant = interpolant
        else:
            interpolant.parameter_positions[0] = now
            interpolant.parameter_positions[1] = now + duration
            interpolant.sample_values[0] = start_time_scale / time_scale
            interpolant.sample_values[1] = end_time_scale / time_scale

        return self

    def stop_warping(self):
        """Stop the warping of the action."""
        self._time_scale_interpolant = None
        return self

    def _update_weight(self, time):
        weight = 0
        if self.enabled:
            weight = self.weight

            weight_interpolant = self._weight_interpolant
            if weight_interpolant is not None:
                interpolant_value = weight_interpolant(time)
                weight *= interpolant_value

                if time > weight_interpolant.parameter_positions[1]:
                    # self._weight_interpolant = None # clear the interpolant
                    self.stop_fading()

                    if interpolant_value == 0:
                        # faded out, disable
                        self.enabled = False

        self._effective_weight = weight
        return weight

    def _update_time_scale(self, time):
        time_scale = 0
        if not self.paused:
            time_scale = self.time_scale

            time_scale_interpolant = self._time_scale_interpolant
            if time_scale_interpolant is not None:
                interpolant_value = time_scale_interpolant(time)
                time_scale *= interpolant_value
                if time > time_scale_interpolant.parameter_positions[1]:
                    # self._time_scale_interpolant = None # clear the interpolant
                    self.stop_warping()

                    if time_scale == 0:
                        # has halted, pause
                        self.paused = True
                    else:
                        # warp done, apply final time scale
                        self.time_scale = time_scale

        self._effective_time_scale = time_scale
        return time_scale

    def _update_time(self, dt):
        time = self.time + dt
        duration = self._clip.duration
        if duration == 0:
            self.enabled = False
            return time

        if time >= duration:
            self.repetitions -= 1
            if self.repetitions <= 0:
                self.enabled = False
                time = duration
            else:
                time = time % duration
                event = Event(type="loop")
                event.action = self
                self._mixer.dispatch_event(event)

        self.time = time
        return time

    def _schedule_fading(self, duration, weight_now, weight_then):
        now = self._mixer._time
        interpolant = self._weight_interpolant
        if interpolant is None:
            times = [now, now + duration]
            values = [weight_now, weight_then]
            interpolant = LinearInterpolant(times, values)
            self._weight_interpolant = interpolant
        else:
            interpolant.parameter_positions[0] = now
            interpolant.parameter_positions[1] = now + duration
            interpolant.sample_values[0] = weight_now
            interpolant.sample_values[1] = weight_then

        return self

    def _update(self, dt):
        if not self.enabled:
            # update ._effective_weight
            self._update_weight(self._mixer._time)
            return

        # apply time scale
        dt = dt * self._update_time_scale(self._mixer._time)
        clip_time = self._update_time(dt)

        weight = self._update_weight(self._mixer._time)

        if weight > 0:
            for track in self._clip.tracks:
                value = track.interpolation(clip_time)
                self._mixer._accumulate(track.target, track.path, value, weight)



import numpy as np

from ._base import Geometry
from .utils import merge


def generate_torso(
    radius_bottom,
    radius_top,
    height,
    radial_segments,
    height_segments,
    theta_start,
    theta_length,
):
    # compute POSITIONS assuming x-y horizontal plane and z up axis

    # radius for each vertex ring from bottom to top
    n_rings = height_segments + 1
    radii = np.linspace(radius_bottom, radius_top, num=n_rings, dtype=np.float32)

    # height for each vertex ring from bottom to top
    half_height = height / 2
    heights = np.linspace(-half_height, half_height, num=n_rings, dtype=np.float32)

    # to enable texture mapping to fully wrap around the cylinder,
    # we can't close the geometry and need a degenerate vertex
    n_vertices = radial_segments + 1

    # xy coordinates on unit circle for a single vertex ring
    theta = np.linspace(
        theta_start, theta_start + theta_length, num=n_vertices, dtype=np.float32
    )
    ring_xy = np.column_stack([np.cos(theta), np.sin(theta)])

    # put all the rings together
    positions = np.empty((n_rings, n_vertices, 3), dtype=np.float32)
    positions[..., :2] = ring_xy[None, ...] * radii[:, None, None]
    positions[..., 2] = heights[:, None]

    # the NORMALS are the same for every ring, so compute for only one ring
    # and then repeat
    slope = (radius_bottom - radius_top) / height
    ring_normals = np.empty(positions.shape[1:], dtype=np.float32)
    ring_normals[..., :2] = ring_xy
    ring_normals[..., 2] = slope
    ring_normals /= np.linalg.norm(ring_normals, axis=-1)[:, None]
    normals = np.empty_like(positions)
    normals[:] = ring_normals[None, ...]

    # the TEXTURE COORDS
    # u maps 0..1 to theta_start..theta_start+theta_length
    # v maps 0..1 to -height/2..height/2
    ring_u = (theta - theta_start) / theta_length
    ring_v = (heights / height) + 0.5
    texcoords = np.empty((n_rings, n_vertices, 2), dtype=np.float32)
    texcoords[..., 0] = ring_u[None, :]
    texcoords[..., 1] = ring_v[:, None]

    # the face INDEX
    # the amount of vertices
    indices = np.arange(n_rings * n_vertices, dtype=np.uint32).reshape(
        (n_rings, n_vertices)
    )
    # for every panel (height_segments, radial_segments) there is a quad (2, 3)
    index = np.empty((height_segments, radial_segments, 2, 3), dtype=np.uint32)
    # create a grid of initial indices for the panels
    index[:, :, 0, 0] = indices[
        np.arange(height_segments)[:, None], np.arange(radial_segments)[None, :]
    ]
    # the remainder of the indices for every panel are relative
    index[:, :, 0, 1] = index[:, :, 0, 0] + 1
    index[:, :, 0, 2] = index[:, :, 0, 0] + n_vertices
    index[:, :, 1, 0] = index[:, :, 0, 0] + n_vertices + 1
    index[:, :, 1, 1] = index[:, :, 1, 0] - 1
    index[:, :, 1, 2] = index[:, :, 1, 0] - n_vertices

    return (
        positions.reshape((-1, 3)),
        normals.reshape((-1, 3)),
        texcoords.reshape((-1, 2)),
        index.flatten(),
    )


def generate_cap(radius, height, radial_segments, theta_start, theta_length, up=True):
    # compute POSITIONS assuming x-y horizontal plane and z up axis

    # to enable texture mapping to fully wrap around the cylinder,
    # we can't close the geometry and need a degenerate vertex
    n_vertices = radial_segments + 1

    # xy coordinates on unit circle for vertex ring
    theta = np.linspace(
        theta_start, theta_start + theta_length, num=n_vertices, dtype=np.float32
    )
    ring_xy = np.column_stack([np.cos(theta), np.sin(theta)])

    # put the vertices together, inserting a center vertex at the start
    positions = np.empty((1 + n_vertices, 3), dtype=np.float32)
    positions[0, :2] = [0.0, 0.0]
    positions[1:, :2] = ring_xy * radius
    positions[..., 2] = height

    # the NORMALS
    normals = np.zeros_like(positions, dtype=np.float32)
    sign = int(up) * 2.0 - 1.0
    normals[..., 2] = sign

    # the TEXTURE COORDS
    # uv etches out a circle from the [0..1, 0..1] range
    # direction is reversed for up=False
    texcoords = np.empty((1 + n_vertices, 2), dtype=np.float32)
    texcoords[0] = [0.5, 0.5]
    texcoords[1:, 0] = ring_xy[:, 0] * 0.5 + 0.5
    texcoords[1:, 1] = ring_xy[:, 1] * 0.5 * sign + 0.5

    # the face INDEX
    indices = np.arange(n_vertices) + 1
    # for every radial segment there is a triangle (3)
    index = np.empty((radial_segments, 3), dtype=np.uint32)
    # create a grid of initial indices for the panels
    index[:, 0] = indices[np.arange(radial_segments)]
    # the remainder of the indices for every panel are relative
    index[:, 1 + int(up)] = n_vertices
    index[:, 2 - int(up)] = index[:, 0] + 1

    return (
        positions,
        normals,
        texcoords,
        index.flatten(),
    )


def cylinder_geometry(
    radius_bottom=1.0,
    radius_top=1.0,
    height=1.0,
    radial_segments=8,
    height_segments=1,
    theta_start=0.0,
    theta_length=np.pi * 2,
    open_ended=False,
):
    """Generate a cylinder or a cylinder segment.

    This function generates a cylinder or a cylinder segment. The cylinder's
    axis runs along the local z-axis, and its midpoint is located at the local
    origin. The cylinder's faces (top and bottom cap) are approximated by
    regular N-sided polygons, with corners on a circle of the given radius.

    Optionally, the cylinder's faces may be replaced with polygon approximations
    of two circle segments. In this case, a cylinder segment will be created,
    and each segment's arc is constructed from N equal-length line segments
    closed by a line along the segments cord.

    Parameters
    ----------
    radius_bottom : float
        The radius at the bottom of the cyliner.
    radius_top : float
        The radius at the top of the cyliner.
    height : float
        The height of the cylinder.
    radial_segments : int
        The number of segments to use when approximating the circle/arc.
    height_segments : int
        The number of evenly spaced segments into which the mantle should be
        split.
    theta_start : float
        The angle (in rad) at which to start the circle segment. Zero points
        into the direction of the local x-axis.
    theta_length : float
        The arc's central angle (in rad). Defaults to a full circle.
    open_ended : bool
        If True, the cylinder's faces are not added and the resulting geometry
        only contains the mantle.

    Returns
    -------
    cylinder : Geometry
        A geometry object representing a cylinder.
        Mathematically, it consists of a set of open orientable manifolds.

    """

    assert radial_segments > 0
    assert height_segments > 0
    assert theta_length != 0.0

    mesh = generate_torso(
        radius_bottom,
        radius_top,
        height,
        radial_segments,
        height_segments,
        theta_start,
        theta_length,
    )

    if not open_ended:
        groups = [mesh]
        if radius_bottom > 0:
            bottom_cap = generate_cap(
                radius_bottom,
                -height / 2,
                radial_segments,
                theta_start,
                theta_length,
                up=False,
            )
            groups.append(bottom_cap)
        if radius_top > 0:
            top_cap = generate_cap(
                radius_top,
                height / 2,
                radial_segments,
                theta_start,
                theta_length,
                up=True,
            )
            groups.append(top_cap)
        mesh = merge(groups)

    positions, normals, texcoords, indices = mesh
    return Geometry(
        indices=indices.reshape((-1, 3)),
        positions=positions,
        normals=normals,
        texcoords=texcoords,
    )


def cone_geometry(
    radius=1.0,
    height=1.0,
    radial_segments=8,
    height_segments=1,
    theta_start=0.0,
    theta_length=np.pi * 2,
    open_ended=False,
):
    """Generate a cone or a cone segment.

    This function generates a cone or a cone segment. The cone's
    axis runs along the local z-axis, and its midpoint is located at the local
    origin.

    The function is a thin wrapper around
    :func:`pygfx.geometries.cylinder_geometry` with ``radius_top = 0.0`` and
    slightly renamed arguments. For details, see the wrapped function.

    Parameters
    ----------
    radius : float
        The radius of the cone's bottom face.
    height : float
        The height of the cone.
    radial_segments : int
        The number of segments to use when approximating the circle/arc.
    height_segments : int
        The number of evenly spaced segments into which the mantle should be
        split.
    theta_start : float
        The angle (in rad) at which to start the circle segment. Zero points
        into the direction of the local x-axis.
    theta_length : float
        The arc's central angle (in rad). Defaults to a full circle.
    open_ended : bool
        If True, the cone's faces are not added and the resulting geometry
        only contains the mantle.

    Returns
    -------
    cylinder : Geometry
        A geometry object representing a cylinder.
        Mathematically, it consists of a set of open orientable manifolds.

    See Also
    --------
    pygfx.cylinder_geometry

    """

    return cylinder_geometry(
        radius_bottom=radius,
        radius_top=0.0,
        height=height,
        radial_segments=radial_segments,
        height_segments=height_segments,
        theta_start=theta_start,
        theta_length=theta_length,
        open_ended=open_ended,
    )



"""pygfx line shader. See line.wgsl for details."""

import wgpu  # only for flags/enums
import numpy as np
import pylinalg as la

from ....utils import array_from_shadertype
from ....resources import Buffer
from ....objects import Line, InstancedLine
from ....materials._line import (
    LineMaterial,
    LineSegmentMaterial,
    LineInfiniteSegmentMaterial,
    LineArrowMaterial,
    LineThinMaterial,
    LineThinSegmentMaterial,
    LineDebugMaterial,
)

from .. import (
    register_wgpu_render_function,
    BaseShader,
    Binding,
    load_wgsl,
    nchannels_from_format,
)


renderer_uniform_type = dict(last_i="i4")


@register_wgpu_render_function(Line, LineMaterial)
class LineShader(BaseShader):
    type = "render"

    def __init__(self, wobject):
        super().__init__(wobject)
        material = wobject.material
        geometry = wobject.geometry

        # Is this an instanced line?
        self["instanced"] = isinstance(wobject, InstancedLine)

        self["line_type"] = "line"
        self["dashing"] = False
        self["thickness_space"] = material.thickness_space
        self["aa"] = material._gfx_effective_aa
        self["loop"] = False
        self["debug"] = False

        # Handle color
        color_mode = str(material.color_mode).split(".")[-1]
        if color_mode == "auto":
            if material.map is not None:
                self["color_mode"] = "vertex_map"
                self["color_buffer_channels"] = 0
            else:
                self["color_mode"] = "uniform"
                self["color_buffer_channels"] = 0
        elif color_mode == "uniform":
            self["color_mode"] = "uniform"
            self["color_buffer_channels"] = 0
        elif color_mode == "vertex":
            nchannels = nchannels_from_format(geometry.colors.format)
            self["color_mode"] = "vertex"
            self["color_buffer_channels"] = nchannels
            if nchannels not in (1, 2, 3, 4):
                raise ValueError(f"Geometry.colors needs 1-4 columns, not {nchannels}")
        elif color_mode == "face":
            nchannels = nchannels_from_format(geometry.colors.format)
            self["color_mode"] = "face"
            self["color_buffer_channels"] = nchannels
            if nchannels not in (1, 2, 3, 4):
                raise ValueError(f"Geometry.colors needs 1-4 columns, not {nchannels}")
        elif color_mode == "vertex_map":
            if material.map is None:
                raise ValueError("Cannot apply colormap is no material.map is set.")
            self["color_mode"] = "vertex_map"
            self["color_buffer_channels"] = 0
        elif color_mode == "face_map":
            if material.map is None:
                raise ValueError("Cannot apply colormap is no material.map is set.")
            self["color_mode"] = "face_map"
            self["color_buffer_channels"] = 0
        else:
            raise RuntimeError(f"Unknown color_mode: '{color_mode}'")

        # Optimization: when the line is opaque, has a uniform color, and no dashing,
        # it can be rendered pretty safely without joins. I *think* this is faster,
        # because a lot of logic related joins becomes simpler. However, the miters
        # result in extra fragments that need to be processed, so we'd need to do
        # some benchmarks to be sure.
        # if (
        #     self["color_mode"] == "uniform"
        #     and not self["dashing"]
        #     and material.alpha_method == "opaque"
        #     and not_using_colors_that_may_have_alpha
        # ):
        #     # self["line_type"] = "quickline"

        # Handle looping. The line_loop_buffer is one larger to enable looping the last point.
        if material.loop:
            self["loop"] = True
            self.line_loop_buffer = Buffer(
                np.zeros((geometry.positions.nitems + 1,), np.uint32)
            )
            self._loop_hash = None
            self.needs_bake_function = True

        # Handle dashing
        if material.dash_pattern:
            # Set dash props
            self["dashing"] = True
            self["dash_pattern"] = tuple(wobject.material.dash_pattern)
            self["dash_count"] = len(wobject.material.dash_pattern) // 2
            # For line segments we can calculate the distance between nodes in the shader.
            # For normal lines, we need a cumulative distance.
            if not isinstance(material, LineSegmentMaterial):
                self.needs_bake_function = True
                self._cumdist_hash = None
                self.line_distance_buffer = Buffer(
                    np.zeros((geometry.positions.nitems,), np.float32)
                )

    def bake_function(self, wobject, camera, logical_size):
        if hasattr(self, "line_loop_buffer"):
            self._bake_line_loops(wobject)
        if hasattr(self, "line_distance_buffer"):
            self._bake_line_distance(wobject, camera, logical_size)

    def _bake_line_loops(self, wobject):
        # Early exit?
        positions_buffer = wobject.geometry.positions
        loop_hash = (id(positions_buffer), positions_buffer.rev)
        if loop_hash == self._loop_hash:
            return
        self._loop_hash = loop_hash

        # Get arrays
        loop_buffer = self.line_loop_buffer
        r_offset, r_size = positions_buffer.draw_range
        positions_array = positions_buffer.data
        loop_array = loop_buffer.data

        # Get indices of points that are nan
        (nan_indices,) = np.where(
            np.isnan(positions_array[r_offset : r_offset + r_size]).any(axis=1)
        )

        is_first = 0x10000000
        is_last = 0x20000000
        is_connector = 0x30000000

        # Mark these indices in the loop array
        loop_array[r_offset : r_offset + r_size] = 0.0
        i1 = r_offset - 1
        i2 = -1
        for i2 in nan_indices:
            n_nodes = i2 - i1 - 1
            if n_nodes >= 3:
                loop_array[i1 + 1] = is_first + n_nodes
                loop_array[i2 - 1] = is_last + n_nodes
                loop_array[i2] = is_connector + n_nodes
            i1 = i2

        # Connect final node to last loop-start. Note that the comparison with i1 and
        # n_nodes makes sure that if the last node is already nan, this step is skipped.
        i2 = r_offset + r_size
        n_nodes = i2 - i1 - 1
        if n_nodes >= 3:
            loop_array[i1 + 1] = is_first + n_nodes
            loop_array[i2 - 1] = is_last + n_nodes
            loop_array[i2] = is_connector + n_nodes

        loop_buffer.update_range(r_offset, r_size)

    def _bake_line_distance(self, wobject, camera, logical_size):
        # Prepare
        positions_buffer = wobject.geometry.positions
        r_offset, r_size = positions_buffer.draw_range

        # Prepare arrays
        positions_array = positions_buffer.data[r_offset : r_offset + r_size]
        distance_array = self.line_distance_buffer.data[r_offset : r_offset + r_size]

        # Get vertices in the appropriate coordinate frame
        if wobject.material.thickness_space == "model":
            # Skip this step if the position data has not changed
            cumdist_hash = (id(positions_buffer), positions_buffer.rev)
            if cumdist_hash == self._cumdist_hash:
                return
            self._cumdist_hash = cumdist_hash
            vertex_array = positions_array
        else:
            # Prep
            finites = np.isfinite(positions_array).all(axis=1)
            has_non_finites = not finites.all()
            if has_non_finites:
                positions_array_sub = positions_array[finites, :]
            else:
                positions_array_sub = positions_array
            # Transform
            if wobject.material.thickness_space == "world":
                vertex_array_sub = la.vec_transform(
                    positions_array_sub, wobject.world.matrix
                )
            else:  # wobject.material.thickness_space == "screen":
                xyz = la.vec_transform(
                    positions_array_sub, camera.camera_matrix @ wobject.world.matrix
                )
                vertex_array_sub = xyz[:, :2] * (0.5 * np.array(logical_size))
            # Fix up
            if has_non_finites:
                vertex_array = np.full((len(positions_array), 2), np.nan, np.float32)
                vertex_array[finites] = vertex_array_sub
            else:
                vertex_array = vertex_array_sub

        # Calculate distances
        distances = np.linalg.norm(vertex_array[1:] - vertex_array[:-1], axis=1)
        distances[~np.isfinite(distances)] = 0.0

        # Store cumulatives
        np.cumsum(distances, out=distance_array[1:])

        # Mark that the data has changed
        self.line_distance_buffer.update_range(r_offset, r_size)

    def get_bindings(self, wobject, shared):
        material = wobject.material
        geometry = wobject.geometry

        positions1 = geometry.positions

        # With vertex buffers, if a shader input is vec4, and the vbo has
        # Nx2, the z and w element will be zero. This works, because for
        # vertex buffers we provide additional information about the
        # striding of the data.
        # With storage buffers (aka SSBO) we just have some bytes that we
        # read from/write to in the shader. This is more free, but it means
        # that the data in the buffer must match with what the shader
        # expects. In addition to that, there's this thing with vec3's which
        # are padded to 16 bytes. So we either have to require our users
        # to provide Nx4 data, or read them as an array of f32.
        # Anyway, extra check here to make sure the data matches!
        if positions1.data is None:
            pass  # assume the user knows that it must be 3D vertices
        elif positions1.data.shape[1] != 3:
            raise ValueError(
                "For rendering (thick) lines, the geometry.positions must be Nx3."
            )

        uniform_buffer = Buffer(
            array_from_shadertype(renderer_uniform_type), force_contiguous=True
        )
        uniform_buffer.data["last_i"] = positions1.nitems - 1

        rbuffer = "buffer/read_only_storage"
        bindings = [
            Binding("u_stdinfo", "buffer/uniform", shared.uniform_buffer),
            Binding("u_wobject", "buffer/uniform", wobject.uniform_buffer),
            Binding("u_material", "buffer/uniform", material.uniform_buffer),
            Binding("u_renderer", "buffer/uniform", uniform_buffer),
            Binding("s_positions", rbuffer, positions1, "VERTEX"),
        ]

        # Per-vertex color, colormap, or a uniform color?
        if self["color_mode"] in ("vertex", "face"):
            bindings.append(Binding("s_colors", rbuffer, geometry.colors, "VERTEX"))
        elif self["color_mode"] in ("vertex_map", "face_map"):
            bindings.append(
                Binding("s_texcoords", rbuffer, geometry.texcoords, "VERTEX")
            )
            bindings.extend(
                self.define_generic_colormap(material.map, geometry.texcoords)
            )

        # Need a buffer for the loop and/or cumdist?
        if hasattr(self, "line_loop_buffer"):
            bindings.append(Binding("s_loop", rbuffer, self.line_loop_buffer, "VERTEX"))
        if hasattr(self, "line_distance_buffer"):
            bindings.append(
                Binding("s_cumdist", rbuffer, self.line_distance_buffer, "VERTEX")
            )

        bindings = {i: b for i, b in enumerate(bindings)}
        self.define_bindings(0, bindings)

        # Instanced lines have an extra storage buffer that we add manually
        bindings1 = {}  # non-auto-generated bindings
        if self["instanced"]:
            bindings1[0] = Binding(
                "s_instance_infos", rbuffer, wobject.instance_buffer, "VERTEX"
            )

        return {
            0: bindings,
            1: bindings1,
        }

    def get_pipeline_info(self, wobject, shared):
        # Cull backfaces so that overlapping faces are not drawn.
        return {
            "primitive_topology": wgpu.PrimitiveTopology.triangle_strip,
            "cull_mode": wgpu.CullMode.none,
        }

    def _get_n(self, positions):
        offset, size = positions.draw_range
        if self["loop"]:
            size += 1
        return offset * 6, size * 6

    def get_render_info(self, wobject, shared):
        # Determine how many vertices are needed
        offset, size = self._get_n(wobject.geometry.positions)
        n_instances = 1
        if self["instanced"]:
            n_instances = wobject.instance_buffer.nitems
        return {
            "indices": (size, n_instances, offset, 0),
        }

    def get_code(self):
        return load_wgsl("line.wgsl")


@register_wgpu_render_function(Line, LineDebugMaterial)
class LineDebugShader(LineShader):
    def __init__(self, wobject):
        super().__init__(wobject)

        self["debug"] = True


@register_wgpu_render_function(Line, LineSegmentMaterial)
class LineSegmentShader(LineShader):
    """This shader is baded on the normal line shader, but it does not draw joins.
    Still needs 6 vertices in for nodes that have a cap on each side.
    """

    def __init__(self, wobject):
        super().__init__(wobject)
        self["line_type"] = "segment"


@register_wgpu_render_function(Line, LineInfiniteSegmentMaterial)
class LineInfiniteSegmentShader(LineShader):
    """Shader to draw infinite line segments. Since the line's ends are always off-screen, there is no need to draw caps."""

    def __init__(self, wobject):
        super().__init__(wobject)
        material = wobject.material
        self["line_type"] = "infsegment"
        self["start_is_infinite"] = material.start_is_infinite
        self["end_is_infinite"] = material.end_is_infinite


@register_wgpu_render_function(Line, LineArrowMaterial)
class LineArrowShader(LineShader):
    """Shader to draw arrows. This shader does not use the caps, so it could be drawn
    with less vertices, but that'd make the code more complex, so for now this is fine.
    """

    def __init__(self, wobject):
        super().__init__(wobject)
        self["line_type"] = "arrow"


# -----  shaders for thin lines


@register_wgpu_render_function(Line, LineThinMaterial)
class ThinLineShader(LineShader):
    type = "render"

    def __init__(self, wobject):
        super().__init__(wobject)
        self["aa"] = False  # no aa with thin lines
        if self["color_mode"] in ("face", "face_map"):
            raise RuntimeError("Face coloring not supported for thin lines.")

    def get_bindings(self, wobject, shared):
        material = wobject.material
        geometry = wobject.geometry

        rbuffer = "buffer/read_only_storage"
        bindings = [
            Binding("u_stdinfo", "buffer/uniform", shared.uniform_buffer),
            Binding("u_wobject", "buffer/uniform", wobject.uniform_buffer),
            Binding("u_material", "buffer/uniform", material.uniform_buffer),
            Binding("s_positions", rbuffer, geometry.positions, "VERTEX"),
        ]

        # Per-vertex color, colormap, or a uniform color?
        if self["color_mode"] == "vertex":
            bindings.append(Binding("s_colors", rbuffer, geometry.colors, "VERTEX"))
        elif self["color_mode"] == "vertex_map":
            bindings.append(
                Binding("s_texcoords", rbuffer, geometry.texcoords, "VERTEX")
            )
            bindings.extend(
                self.define_generic_colormap(material.map, geometry.texcoords)
            )

        bindings = {i: b for i, b in enumerate(bindings)}
        self.define_bindings(0, bindings)

        return {
            0: bindings,
        }

    def get_pipeline_info(self, wobject, shared):
        return {
            "primitive_topology": wgpu.PrimitiveTopology.line_strip,
            "cull_mode": wgpu.CullMode.none,
        }

    def get_render_info(self, wobject, shared):
        offset, size = wobject.geometry.positions.draw_range
        return {
            "indices": (size, 1, offset, 0),
        }

    def get_code(self):
        return """//wgsl

        {$ include 'pygfx.std.wgsl' $}

        struct VertexInput {
            @builtin(vertex_index) index : u32,
        };

        @vertex
        fn vs_main(in: VertexInput) -> Varyings {

            let i0 = i32(in.index);

            let raw_pos = load_s_positions(i0);
            let wpos = u_wobject.world_transform * vec4<f32>(raw_pos.xyz, 1.0);
            let npos = u_stdinfo.projection_transform * u_stdinfo.cam_transform * wpos;

            var varyings: Varyings;
            varyings.position = vec4<f32>(npos);
            varyings.world_pos = vec3<f32>(ndc_to_world_pos(npos));

            // per-vertex or per-face coloring
            $$ if color_mode == 'vertex'
                let color_index = i0;
                $$ if color_buffer_channels == 1
                    let cvalue = load_s_colors(color_index);
                    varyings.color = vec4<f32>(cvalue, cvalue, cvalue, 1.0);
                $$ elif color_buffer_channels == 2
                    let cvalue = load_s_colors(color_index);
                    varyings.color = vec4<f32>(cvalue.r, cvalue.r, cvalue.r, cvalue.g);
                $$ elif color_buffer_channels == 3
                    varyings.color = vec4<f32>(load_s_colors(color_index), 1.0);
                $$ elif color_buffer_channels == 4
                    varyings.color = vec4<f32>(load_s_colors(color_index));
                $$ endif
            $$ endif

            // Set texture coords
            let tex_coord_index = i0;
            $$ if colormap_dim == '1d'
            varyings.texcoord = f32(load_s_texcoords(tex_coord_index));
            $$ elif colormap_dim == '2d'
            varyings.texcoord = vec2<f32>(load_s_texcoords(tex_coord_index));
            $$ elif colormap_dim == '3d'
            varyings.texcoord = vec3<f32>(load_s_texcoords(tex_coord_index));
            $$ endif

            return varyings;
        }

        @fragment
        fn fs_main(varyings: Varyings) -> FragmentOutput {
            {$ include 'pygfx.clipping_planes.wgsl' $}

            $$ if color_mode == 'vertex'
                let color = varyings.color;
            $$ elif color_mode == 'vertex_map'
                let color = sample_colormap(varyings.texcoord);
            $$ else
                let color = u_material.color;
            $$ endif

            let physical_color = srgb2physical(color.rgb);
            let opacity = color.a * u_material.opacity;
            let out_color = vec4<f32>(physical_color, opacity);

            do_alpha_test(opacity);

            var out: FragmentOutput;
            out.color = out_color;
            return out;
        }
        """


@register_wgpu_render_function(Line, LineThinSegmentMaterial)
class ThinLineSegmentShader(ThinLineShader):
    def get_pipeline_info(self, wobject, shared):
        return {
            "primitive_topology": wgpu.PrimitiveTopology.line_list,
            "cull_mode": wgpu.CullMode.none,
        }



"""
.. currentmodule:: pygfx.renderers.wgpu


General functions and classes to operate more closely with the wgpu backend.

.. autosummary::
    :toctree: _autosummary/renderers/wgpu
    :template: ../_templates/custom_layout.rst

    print_wgpu_report
    select_power_preference
    select_adapter
    enable_wgpu_features
    set_wgpu_limits
    WgpuRenderer
    register_wgpu_render_function
    get_shared
    Shared


Classes and functions required to implement custom shaders:

.. autosummary::
    :toctree: _autosummary/renderers/wgpu
    :template: ../_templates/custom_layout.rst

    ShaderInterface
    BaseShader
    Binding
    GfxSampler
    GfxTextureView
    register_wgsl_loader
    load_wgsl


Lower level functions that may or may not be needed in custom shaders:

.. autosummary::
    :toctree: _autosummary/renderers/wgpu
    :template: ../_templates/custom_layout.rst

    nchannels_from_format
    to_index_format
    to_vertex_format
    to_texture_format

"""

# ruff: noqa: F401

# Import stuff that people need who create custom shaders, so they can import from pygfx.renderers.wgpu
from .engine.utils import (
    register_wgpu_render_function,
    nchannels_from_format,
    to_index_format,
    to_vertex_format,
    to_texture_format,
    GfxSampler,
    GfxTextureView,
)
from .engine.shared import (
    Shared,
    get_shared,
    print_wgpu_report,
    select_power_preference,
    select_adapter,
    enable_wgpu_features,
    set_wgpu_limits,
)
from .engine.renderer import WgpuRenderer
from .engine.effectpasses import (
    EffectPass,
    CopyPass,
    PPAAPass,
    FXAAPass,
    DDAAPass,
    NoisePass,
    FogPass,
    DepthPass,
)
from .engine.pipeline import Binding

# Shader classes
from .wgsl import load_wgsl
from .shader import ShaderInterface, BaseShader, register_wgsl_loader

# Load shaders submodules, so that all our builtin shaders become a vailable
from . import shaders



"""
Utility functions for Pygfx.

.. currentmodule:: pygfx.utils

.. autosummary::
    :toctree: utils/
    :template: ../_templates/custom_layout.rst

    Color
    load.load_mesh
    load.load_meshes
    load.load_scene
    show.show
    show.Display
    viewport.Viewport
    text.FontManager
    enums
    cm

Transform classes
-----------------

These classes are used internally to create an intuitive API to transform
WorldObjects (`obj.world` and `obj.local`). They are listed here for
completeness and you will likely never have to instantiate them directly.

.. autosummary::
    :toctree: utils/
    :template: ../_templates/custom_layout.rst

    transform.AffineBase
    transform.AffineTransform
    transform.RecursiveTransform

"""

import os
import types
import logging
import inspect

import numpy as np

from .color import Color  # noqa: F401
from . import enums  # noqa: F401

from ._dirs import get_resources_dir, get_cache_dir  # noqa: F401

logger = logging.getLogger("pygfx")


def _set_log_level():
    # Set default level
    logger.setLevel(logging.WARN)
    # Set user-specified level
    level = os.getenv("PYGFX_LOG_LEVEL", "")
    if level:
        try:
            if level.isnumeric():
                logger.setLevel(int(level))
            else:
                logger.setLevel(level.upper())
        except Exception:
            logger.warning(f"Invalid pygfx log level: {level}")


_set_log_level()


def array_from_shadertype(shadertype, count=None):
    """Get a numpy array object from a dict shadertype.

    The fields are re-ordered and padded as necessary to fulfil alignment rules.
    See https://www.w3.org/TR/WGSL/#structure-layout-rules

    A note on matrices (e.g. "4x4xf4" or "2x3xf4"): from the perspective of Python,
    the matrices in a uniform array are transposed, and nx3 matrices are actually nx4.

    To deal with this, setting a matrix in a uniform goes like this:

        uniform.data["a_matrix"] = numpy_array.T
        uniform.data["an_nx3_matrix"][:, :3] = numpy_array.T

    The reason is that WGSL matrices are column-major, while Numpy arrays are
    row-major (i.e. C-order) by default. Although numpy does support
    column-major arrays (F-order), it looks like we cannot apply this to the
    sub-arrays in the uniform struct. And for the nx3 matrices, WGSL's alignment
    constraint requires padding for each column in an nx3 matrix, so it takes up
    the same space as an nx4 matrix.

    params:
        shadertype: dict
            A dict containing the shadertype.
        count: None or int
            If count is not None, array has a shape of (count, ),
            Indicates that the corresponding "Buffer" is an array of struct.
    """
    assert isinstance(shadertype, dict)
    assert count is None or count > 0

    # Note: we have yet to add f2 (16 bit float)
    primitives = {
        "i4": "int32",
        "u4": "uint32",
        "f4": "float32",
    }

    class Field:
        __slots__ = [
            "align",
            "name",
            "primitive",
            "shape",
            "size",
        ]

        def __init__(self, name, format):
            self.name = name
            if format[-2:] not in primitives:
                raise RuntimeError(
                    f"Values in a uniform must have a 32bit primitive type, not {format}"
                )
            self.primitive = primitives[format[-2:]]
            primitive_size = 4  # no support for f2 yet
            # Get shape, excluding array part
            arraystr, _, shapestr = format[:-2].rpartition("*")
            shape = [int(i) for i in shapestr.split("x") if i]
            shape = shape or [1]
            # The alignment is based on the last element of te (original) shape
            self.align = shape[-1] * primitive_size
            # Snap align to factors of 2.
            # 3xf2 -> 6 -> 8
            # 3x3xf4 -> 12 -> 16
            for ref_align in [2, 4, 8, 16]:
                if self.align <= ref_align:
                    self.align = ref_align
                    break
            # Deal with nx3 matrices; they are nx4 internally
            if len(shape) > 1 and shape[-1] == 3:
                mat3_names.append(name)
                shape[-1] = 4
            # Include array size in shape
            if arraystr:
                array_names.append(name)
                shape.insert(0, int(arraystr))
            # Calculate size, the number of bytes that this field occupies
            self.size = int(np.prod(shape)) * primitive_size
            # Convert shape to tuple for the numpy sub array.
            # We do NOT reverse the shape, see function docstring.
            self.shape = () if (len(shape) == 1 and shape[0] == 1) else tuple(shape)

        def use(self):
            result = self.name, self.primitive, self.shape
            self.name += "-already used"  # break when used twice
            return result

    pad_index = 0
    array_names = []
    mat3_names = []
    fields_per_align = {16: [], 8: [], 4: [], 2: []}
    packed_fields = []  # what fields fill up space to fix alignment

    # The definition of the dtype of the structured array, as a list of tuples: (name, primitive-dtype, shape)
    # We could also use the dict flavor, setting explicit offsets for each field instead of stub padding fields,
    # but its also an advantage during debugging that the padding is explicit/visible.
    dtype_fields = []

    # The size/alignment of a struct is the max alignment of its members.
    # But when a uniform buffer is an array of structs, the striding of the array must be a multiple of 16 (std140-like rules).
    struct_alignment = 0
    if count is not None:
        struct_alignment = 16

    # Collect fields per alignment, so we can process from big to small to avoid padding.
    for name, format in shadertype.items():
        field = Field(name, format)
        fields_per_align[field.align].append(field)
        struct_alignment = max(struct_alignment, field.align)

    def fill_bytes(ref_align, nbytes):
        # This function only gets called when using a vec3<x32> or vec3<f16>. Since the
        # fields are sorted by alignment (big to small) this only applies to types that
        # align weirdly. But mot for the nx3 matrices, because they have intrinsic padding.
        nonlocal pad_index
        j = ref_align - nbytes
        # See if we can fill the gap with other fields
        for align in [4, 2]:
            if align >= ref_align or j % align > 0:
                continue
            fields = fields_per_align[align]
            checked_all = False
            while nbytes >= align and not checked_all:
                selected_field = None
                for field in fields:
                    if field.size <= nbytes:
                        selected_field = field
                        packed_fields.append(field.name)
                        dtype_fields.append(field.use())
                        nbytes -= field.size
                        j += field.size
                        break
                if selected_field:
                    fields.remove(selected_field)
                else:
                    checked_all = True
        # Fill remaining space with padding
        if nbytes:
            pad_index += 1
            dtype_fields.append((f"__padding{pad_index}", "uint8", (need_bytes,)))

    # Process the fields, from big to small
    i = 0  # bytes processed
    for ref_align in [16, 8, 4, 2]:
        for field in fields_per_align[ref_align]:
            too_many_bytes = i % field.align
            if too_many_bytes:
                need_bytes = field.align - too_many_bytes
                assert need_bytes < 16
                assert ref_align in (16, 8)
                fill_bytes(ref_align, need_bytes)
                i += need_bytes
            # Add the field itself
            dtype_fields.append(field.use())
            i += field.size

    # Add padding to the struct
    too_many_bytes = i % struct_alignment
    if too_many_bytes:
        need_bytes = struct_alignment - too_many_bytes
        pad_index += 1
        dtype_fields.append((f"__padding{pad_index}", "uint8", (need_bytes,)))
        i += need_bytes

    # Nice for debugging:
    # if packed_fields or pad_index > 1:
    #     print("fields:")
    #     for field in dtype_fields:
    #         print("   ", field)

    # Add meta fields (zero bytes)
    # This isn't particularly pretty, but this way our metadata is attached
    # to the dtype without affecting its size.
    if array_names:
        dtype_fields.append(
            ("__meta_array_names__" + "__".join(array_names), "uint8", (0,))
        )
    if mat3_names:
        dtype_fields.append(
            ("__meta_mat3_names__" + "__".join(mat3_names), "uint8", (0,))
        )

    # Create a scalar of this type
    if count is not None:
        uniform_data = np.zeros((count,), dtype=dtype_fields)
    else:
        uniform_data = np.zeros((), dtype=dtype_fields)

    # If this fails we did something wrong above
    assert uniform_data.nbytes % struct_alignment == 0

    return uniform_data


def unpack_bitfield(packed, **bit_counts):
    """Unpack values from an uint64 bitfield."""
    values = {}
    for key, bits in bit_counts.items():
        mask = 2**bits - 1
        values[key] = packed & mask
        packed = packed >> bits
    return values


def normals_from_vertices(rr, tris):
    """Efficiently compute vertex normals for a triangulated surface."""
    # This code was taken from Vispy
    # ensure highest precision for our summation/vectorization "trick"
    rr = rr[:, :3].astype(np.float64)
    tris = tris.reshape(-1, 3)
    # first, compute triangle normals
    r1 = rr[tris[:, 0], :]
    r2 = rr[tris[:, 1], :]
    r3 = rr[tris[:, 2], :]
    tri_nn = np.cross((r2 - r1), (r3 - r1))

    # Triangle normals and areas
    size = np.sqrt(np.sum(tri_nn * tri_nn, axis=1))
    size[size == 0] = 1.0  # prevent ugly divide-by-zero
    tri_nn /= size[:, np.newaxis]

    npts = len(rr)

    # the following code replaces this, but is faster (vectorized):
    #
    # for p, verts in enumerate(tris):
    #     nn[verts, :] += tri_nn[p, :]
    #
    nn = np.zeros((npts, 3))
    for verts in tris.T:  # note this only loops 3x (number of verts per tri)
        for idx in range(3):  # x, y, z
            nn[:, idx] += np.bincount(
                verts.astype(np.int32), tri_nn[:, idx], minlength=npts
            )
    size = np.sqrt(np.sum(nn * nn, axis=1))
    size[size == 0] = 1.0  # prevent ugly divide-by-zero
    nn /= size[:, np.newaxis]
    return nn.astype(np.float32)


def assert_type(name, value, *classes):
    allow_none = False
    if classes[0] is None:
        if value is None:
            return
        allow_none = True
        classes = classes[1:]

    if not isinstance(value, classes):
        # Get traceback object to point of the frame of interest
        f = inspect.currentframe()
        f = f.f_back
        if name:
            # Step back to calling code
            f = f.f_back
            # If this is a constructor that has name as a (kw) argument, take another step back
            if f.f_code.co_name == "__init__" and name in f.f_code.co_varnames:
                f = f.f_back
        tb = types.TracebackType(None, f, f.f_lasti, f.f_lineno)

        # Build error message
        msg = "Expected"
        if name:
            msg += f" '{name}' to be"
        class_strings = [cls.__name__ for cls in classes]
        msg += f" an instance of {' | '.join(class_strings)}"
        if allow_none:
            msg += " or None"
        valuestr = value.__class__.__name__
        msg += f", but got {valuestr} object."

        # Raise message with alt traceback
        raise TypeError(msg).with_traceback(tb) from None


class ReadOnlyDict(dict):
    """A read-only dict, for storing structured data that can be hashed."""

    __slots__ = ["_hash"]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # Calculate hash in a way that requires any value to also be hashable
        parts = []
        for k in sorted(self.keys()):
            v = self[k]
            parts.append(str(hash(k)))
            parts.append(str(hash(v)))
        self._hash = hash(" ".join(parts))

    def __setitem__(self, *args, **kwargs):
        raise TypeError("Cannot modify ReadOnlyDict")

    def __delitem__(self, *args, **kwargs):
        raise TypeError("Cannot modify ReadOnlyDict")

    def clear(self, *args, **kwargs):
        raise TypeError("Cannot modify ReadOnlyDict")

    def pop(self, *args, **kwargs):
        raise TypeError("Cannot modify ReadOnlyDict")

    def popitem(self, *args, **kwargs):
        raise TypeError("Cannot modify ReadOnlyDict")

    def setdefault(self, *args, **kwargs):
        raise TypeError("Cannot modify ReadOnlyDict")

    def update(self, *args, **kwargs):
        raise TypeError("Cannot modify ReadOnlyDict")

    def __hash__(self):
        return self._hash
